"""DataFrame checks."""

import pandas as pd
import pandera.extensions as extensions

from bfb_delivery.lib.constants import CircuitColumns, Columns, IntermediateColumns

# NOTE: Registering as dataframe checks instead of field checks includes the columns in the
# error message, whereas groupby field checks do not.
# NOTE: There may be better a way to mock an RDB structure (spin up temp DB), but this works.


@extensions.register_check_method(statistics=["flag"])
def at_least_one_in_group_sheet_plan(df: pd.DataFrame, flag: bool) -> bool:
    """Check that at least one value in a group is not null or empty."""
    return (
        at_least_one_in_group(
            df=df,
            group_col=IntermediateColumns.DRIVER_SHEET_NAME,
            at_least_one_col=CircuitColumns.PLAN,
        )
        if flag
        else True
    )


@extensions.register_check_method(statistics=["flag"])
def at_least_one_in_group_route_sheet(df: pd.DataFrame, flag: bool) -> bool:
    """Check that at least one value in a group is not null or empty."""
    return (
        at_least_one_in_group(
            df=df,
            group_col=CircuitColumns.ROUTE,
            at_least_one_col=IntermediateColumns.DRIVER_SHEET_NAME,
        )
        if flag
        else True
    )


@extensions.register_check_method(statistics=["flag"])
def at_least_one_in_group_sheet_route(df: pd.DataFrame, flag: bool) -> bool:
    """Check that at least one value in a group is not null or empty."""
    return (
        at_least_one_in_group(
            df=df,
            group_col=IntermediateColumns.DRIVER_SHEET_NAME,
            at_least_one_col=CircuitColumns.ROUTE,
        )
        if flag
        else True
    )


@extensions.register_check_method(statistics=["flag"])
def at_least_one_in_group_sheet_stop(df: pd.DataFrame, flag: bool) -> bool:
    """Check that at least one value in a group is not null or empty."""
    return (
        at_least_one_in_group(
            df=df,
            group_col=IntermediateColumns.DRIVER_SHEET_NAME,
            at_least_one_col=Columns.STOP_NO,
        )
        if flag
        else True
    )


@extensions.register_check_method(statistics=["group_col", "at_least_one_col"])
def at_least_one_in_group(df: pd.DataFrame, group_col: str, at_least_one_col: str) -> bool:
    """Check that at least one value in a group is not null or empty."""
    return all(df.groupby(group_col)[at_least_one_col].count() > 0)


@extensions.register_check_method(statistics=["group_col", "contiguous_col", "start_idx"])
def contiguous_group(
    df: pd.DataFrame, group_col: str, contiguous_col: str, start_idx: int
) -> bool:
    """Assert that values are contiguous in each group."""
    return all(
        sorted(vals) == list(range(start_idx, len(vals) + start_idx))
        for _, vals in df.groupby(group_col)[contiguous_col]
    )


@extensions.register_check_method(statistics=["col_a", "col_b"])
def equal_cols(df: pd.DataFrame, col_a: str, col_b: str) -> bool:
    """Assert that plan titles are the same as route titles."""
    return all(df[col_a] == df[col_b])


@extensions.register_check_method(statistics=["cols"])
def increasing_by(df: pd.DataFrame, cols: list[str]) -> bool:
    """Assert that a DataFrame is sorted by columns."""

    def increasing_groupby(df: pd.DataFrame, col_a: str, col_b: str) -> bool:
        return df[col_a].is_monotonic_increasing and all(
            series.is_monotonic_increasing for _, series in df.groupby(col_a)[col_b]
        )

    return all(
        increasing_groupby(df, col_a=cols[i], col_b=cols[i + 1]) for i in range(len(cols) - 1)
    )


@extensions.register_check_method(statistics=["flag"])
def address1_in_address(df: pd.DataFrame, flag: bool) -> bool:
    """Check that a address line one in address."""
    return (
        item_in_dict_col(
            df=df, col_name=CircuitColumns.ADDRESS, item_name=CircuitColumns.ADDRESS_LINE_1
        )
        if flag
        else True
    )


@extensions.register_check_method(statistics=["flag"])
def address2_in_address(df: pd.DataFrame, flag: bool) -> bool:
    """Check that a address line two in address."""
    return (
        item_in_dict_col(
            df=df, col_name=CircuitColumns.ADDRESS, item_name=CircuitColumns.ADDRESS_LINE_2
        )
        if flag
        else True
    )


@extensions.register_check_method(statistics=["col_name", "item_name"])
def item_in_dict_col(df: pd.DataFrame, col_name: str, item_name: str) -> bool:
    """Check that a dictionary field has an item in it."""
    return all(item_name in val.keys() for val in df[col_name])


@extensions.register_check_method(statistics=["many_col", "one_col"])
def many_to_one(df: pd.DataFrame, many_col: str, one_col: str) -> bool:
    """Assert that a column has a many-to-one relationship with another column."""
    return df.groupby(many_col)[one_col].nunique().eq(1).all()


@extensions.register_check_method(statistics=["flag"])
def one_to_one_route_sheet(df: pd.DataFrame, flag: bool) -> bool:
    """Assert that columns have a 1:1 relationship."""
    return (
        one_to_one(
            df=df, col_a=CircuitColumns.ROUTE, col_b=IntermediateColumns.DRIVER_SHEET_NAME
        )
        if flag
        else True
    )


@extensions.register_check_method(statistics=["col_a", "col_b"])
def one_to_one(df: pd.DataFrame, col_a: str, col_b: str) -> bool:
    """Assert that columns have a 1:1 relationship."""
    return (
        df.groupby(col_a)[col_b].nunique().eq(1).all()
        and df.groupby(col_b)[col_a].nunique().eq(1).all()
    )


@extensions.register_check_method(statistics=["group_col", "unique_col"])
def unique_group(df: pd.DataFrame, group_col: str, unique_col: str) -> bool:
    """Assert that values are unique in each group."""
    return all(len(vals) == vals.nunique() for _, vals in df.groupby(group_col)[unique_col])



"""Write routes to Circuit."""

import logging
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any

import pandas as pd
import pandera as pa
from pandera.typing import DataFrame
from typeguard import typechecked

from comb_utils import concat_response_pages, get_responses

from bfb_delivery.lib import errors, schema
from bfb_delivery.lib.constants import (
    CIRCUIT_DATE_FORMAT,
    CIRCUIT_DRIVERS_URL,
    MANIFEST_DATE_FORMAT,
    CircuitColumns,
    Columns,
    DocStrings,
    IntermediateColumns,
    RateLimits,
)
from bfb_delivery.lib.dispatch.api_callers import (
    OptimizationChecker,
    OptimizationLauncher,
    PagedResponseGetterBFB,
    PlanDeleter,
    PlanDistributor,
    PlanInitializer,
    StopUploader,
)
from bfb_delivery.lib.dispatch.read_circuit import get_route_files
from bfb_delivery.lib.formatting.sheet_shaping import create_manifests, split_chunked_route
from bfb_delivery.lib.schema.utils import schema_error_handler
from bfb_delivery.lib.utils import get_friday

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@typechecked
def build_routes_from_chunked(  # noqa: D103
    input_path: str,
    output_dir: str,
    start_date: str,
    no_distribute: bool,
    verbose: bool,
    extra_notes_file: str,
) -> Path:
    start_date = start_date or get_friday(fmt=CIRCUIT_DATE_FORMAT)

    input_path = str(Path(input_path).resolve())
    output_dir = output_dir if output_dir else f"./deliveries_{start_date}"
    output_dir = str(Path(output_dir).resolve())
    Path(output_dir).mkdir(exist_ok=True)
    split_chunked_output_dir = Path(output_dir) / "split_chunked"
    split_chunked_output_dir.mkdir(exist_ok=True)

    split_chunked_workbook_fp = split_chunked_route(
        input_path=input_path,
        output_dir=split_chunked_output_dir,
        output_filename="",
        n_books=1,
        book_one_drivers_file="",
        date=datetime.strptime(start_date, CIRCUIT_DATE_FORMAT).strftime(
            MANIFEST_DATE_FORMAT
        ),
    )[0]

    plan_df = upload_split_chunked(
        split_chunked_workbook_fp=split_chunked_workbook_fp,
        output_dir=Path(output_dir),
        start_date=start_date,
        no_distribute=no_distribute,
        verbose=verbose,
    )

    successful_plans = plan_df[plan_df[IntermediateColumns.OPTIMIZED] == True][  # noqa: E712
        IntermediateColumns.PLAN_ID
    ].to_list()

    circuit_output_dir = get_route_files(
        start_date=start_date,
        end_date=start_date,
        plan_ids=successful_plans,
        output_dir=output_dir,
        verbose=verbose,
    )

    final_manifest_path = create_manifests(
        input_dir=circuit_output_dir,
        output_dir=output_dir,
        output_filename="",
        extra_notes_file=str(extra_notes_file) if extra_notes_file else "",
    )

    # TODO: Validate that input data matches ouput data.
    # (Ignore skipped routes, and account for imputation, caps, etc.)
    # https://github.com/crickets-and-comb/bfb_delivery/issues/62

    return final_manifest_path


build_routes_from_chunked.__doc__ = DocStrings.BUILD_ROUTES_FROM_CHUNKED.api_docstring


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def upload_split_chunked(
    split_chunked_workbook_fp: Path,
    output_dir: Path,
    start_date: str,
    no_distribute: bool,
    verbose: bool,
) -> DataFrame[schema.PlansUploadSplitChunkedOut]:
    """Upload, optimize, and distribute a split chunked Excel workbook of routes to Circuit.

    The workbook contains multiple sheets, one per route. Each sheet is named after the driver
    with the date. The columns are:

        - Name
        - Address
        - Phone
        - Email
        - Notes
        - Order Count
        - Product Type
        - Neighborhood

    Args:
        split_chunked_workbook_fp: The file path to the split chunked workbook.
        output_dir: A place to put intermediate files.
        start_date: The date to start the routes, as "YYYY-MM-DD".
        no_distribute: To skip distributing the routes after optimizing.
        verbose: Whether to print verbose output.

    Returns:
        A DataFrame with the plan IDs and driver IDs for each sheet,
            along with date and statuses at each step.
    """
    plan_output_dir = output_dir / "plans"
    plan_output_dir.mkdir(exist_ok=True)
    plan_df_path = plan_output_dir / "plans.csv"
    stops_df_path = plan_output_dir / "stops.csv"

    stops_df = _create_stops_df(
        split_chunked_workbook_fp=split_chunked_workbook_fp, stops_df_path=stops_df_path
    )

    plan_df = _create_plans(
        stops_df=stops_df, start_date=start_date, plan_df_path=plan_df_path, verbose=verbose
    )
    plan_df.to_csv(plan_df_path, index=False)

    plan_df = _upload_stops(stops_df=stops_df, plan_df=plan_df, verbose=verbose)
    plan_df.to_csv(plan_df_path, index=False)

    plan_df = _optimize_routes(plan_df=plan_df, verbose=verbose)
    plan_df.to_csv(plan_df_path, index=False)

    if not no_distribute:
        plan_df = _distribute_routes(plan_df=plan_df, verbose=verbose)
    else:
        plan_df[CircuitColumns.DISTRIBUTED] = False

    plan_df[IntermediateColumns.START_DATE] = start_date
    plan_df.to_csv(plan_df_path, index=False)

    _print_report(plan_df=plan_df, no_distribute=no_distribute)

    return plan_df


@typechecked
def delete_plans(plan_ids: list[str], plan_df_fp: str) -> list[str]:
    """Delete plans from Circuit.

    Args:
        plan_ids: The plan IDs to delete.
        plan_df_fp: The file path to a dataframe with plan IDs to be deleted
            in column 'plan_id'.

    Returns:
        The plan IDs that were (to be) deleted.

    Raises:
        ValueError: If both plan_ids and plan_df_fp are provided.
        ValueError: If neither plan_ids nor plan_df_fp are provided.
        RuntimeError: If there are errors deleting plans.
    """
    if plan_ids and plan_df_fp:
        raise ValueError("Please provide either plan_ids or plan_ids_fp, not both.")
    if not plan_ids and not plan_df_fp:
        raise ValueError("Please provide either plan_ids or plan_ids_fp.")

    if plan_df_fp:
        plan_df = pd.read_csv(plan_df_fp)
        plan_ids = plan_df[IntermediateColumns.PLAN_ID].to_list()

    logger.info(f"Deleting plans: {plan_ids} ...")

    errors = {}
    for plan_id in plan_ids:
        try:
            deleted = delete_plan(plan_id=plan_id)
        except Exception as e:
            errors[plan_id] = e
            logger.error(f"Error deleting plan {plan_id}.")
        else:
            logger.info(f"Plan {plan_id} deleted: {deleted}")

    logger.info("Finished deleting plans.")

    if errors:
        raise RuntimeError(f"Errors deleting plans:\n{errors}")

    return plan_ids


@typechecked
def delete_plan(plan_id: str) -> bool:
    """Delete a plan from Circuit.

    Args:
        plan_id: The plan ID to be deleted.

    Returns:
        Whether the plan was deleted. (Should always be True if no errors.)
    """
    deleter = PlanDeleter(plan_id=plan_id)
    deleter.call_api()

    return deleter.deletion


@typechecked
def _create_stops_df(split_chunked_workbook_fp: Path, stops_df_path: Path) -> pd.DataFrame:
    stops_dfs = []
    with pd.ExcelFile(split_chunked_workbook_fp) as workbook:
        for sheet in workbook.sheet_names:
            df = workbook.parse(sheet)
            df[IntermediateColumns.SHEET_NAME] = str(sheet)
            stops_dfs.append(df)
    stops_df = pd.concat(stops_dfs).reset_index(drop=True)
    stops_df = stops_df.fillna("")
    stops_df.to_csv(stops_df_path, index=False)

    return stops_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _create_plans(
    stops_df: DataFrame[schema.Stops], start_date: str, plan_df_path: Path, verbose: bool
) -> DataFrame[schema.PlansCreatePlansOut]:
    """Create a Circuit plan for each route.

    Args:
        stops_df: The long DataFrame with all the routes.
        start_date: The date to start the routes, as "YYYY-MM-DD".
        plan_df_path: The file path to save the plan DataFrame in case of error.
            Helps to automate cleanup.
        verbose: Whether to print verbose output.
    """
    plan_df = _assign_drivers_to_plans(stops_df=stops_df)
    try:
        plan_df = _initialize_plans(plan_df=plan_df, start_date=start_date, verbose=verbose)
    except Exception as e:
        plan_df.to_csv(plan_df_path, index=False)
        raise e

    return plan_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _upload_stops(
    stops_df: DataFrame[schema.Stops],
    plan_df: DataFrame[schema.PlansUploadStopsIn],
    verbose: bool,
) -> DataFrame[schema.PlansUploadStopsOut]:
    """Upload stops for each route.

    Will skip plans marked as not initialized.

    Args:
        stops_df: The long DataFrame with all the routes.
        plan_df: The DataFrame with the plan IDs and driver IDs for each sheet.
        verbose: Whether to print verbose output.

    Returns:
        A dataframe of the plans with a column for whether stops were uploaded successfully.
    """
    plan_stops = _build_plan_stops(
        stops_df=stops_df,
        plan_df=plan_df[plan_df[IntermediateColumns.INITIALIZED] == True],  # noqa: E712
    )

    logger.info(
        f"Uploading stops. Allow {RateLimits.BATCH_STOP_IMPORT_SECONDS}+ seconds per plan ..."
    )
    uploaded_stops = {}
    stop_id_count = 0
    errors = {}
    for plan_id, stop_arrays in plan_stops.items():
        plan_title = plan_df[plan_df[IntermediateColumns.PLAN_ID] == plan_id][
            IntermediateColumns.ROUTE_TITLE
        ].values[0]

        if verbose:
            logger.info(f"Uploading stops for {plan_title} ({plan_id}) ...")

        for stop_array in stop_arrays:
            stop_uploader = StopUploader(
                plan_id=plan_id, plan_title=plan_title, stop_array=stop_array
            )
            try:
                stop_uploader.call_api()
            except Exception as e:
                logger.error(f"Error uploading stops for {plan_title} ({plan_id}):\n{e}")
                if plan_id not in errors:
                    errors[plan_title] = [e]
                else:
                    errors[plan_title].append(e)
            else:
                uploaded_stops[plan_title] = stop_uploader.stop_ids
                stop_id_count += len(stop_uploader.stop_ids)

    logger.info(
        f"Finished uploading stops. Uploaded {stop_id_count} stops for "
        f"{len(uploaded_stops)} plans."
    )

    plan_df[IntermediateColumns.STOPS_UPLOADED] = False
    plan_df.loc[
        (plan_df[IntermediateColumns.PLAN_ID].isin(plan_stops.keys()))
        & ~(plan_df[IntermediateColumns.ROUTE_TITLE].isin(errors.keys())),
        IntermediateColumns.STOPS_UPLOADED,
    ] = True
    if errors:
        logger.warning(f"Errors uploading stops:\n{errors}")

    return plan_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _optimize_routes(
    plan_df: DataFrame[schema.PlansOptimizeRoutesIn], verbose: bool
) -> DataFrame[schema.PlansOptimizeRoutesOut]:
    """Optimize the routes.

    Will skip plans marked as without stops uploaded.

    Args:
        plan_df: The DataFrame with the plan IDs and driver IDs for each sheet.
        verbose: Whether to print verbose output.

    Returns:
        A DataFrame of the plans with a column for whether the routes were optimized.
    """
    logger.info(
        "Initializing route optimizations. "
        f"Allow {RateLimits.OPTIMIZATION_PER_SECOND}+ seconds per plan ..."
    )
    plan_ids = plan_df[plan_df[IntermediateColumns.STOPS_UPLOADED] == True][  # noqa: E712
        IntermediateColumns.PLAN_ID
    ].to_list()
    optimizations = {}

    errors = {}
    for plan_id in plan_ids:
        plan_title = plan_df[plan_df[IntermediateColumns.PLAN_ID] == plan_id][
            IntermediateColumns.ROUTE_TITLE
        ].values[0]
        if verbose:
            logger.info(f"Optimizing route for {plan_title} ({plan_id}) ...")

        optimization = OptimizationLauncher(plan_id=plan_id, plan_title=plan_title)
        try:
            optimization.call_api()
        except Exception as e:
            logger.error(f"Error launching optimization for {plan_title} ({plan_id}):\n{e}")
            errors[plan_title] = e

        else:
            optimizations[plan_id] = optimization.operation_id
            if verbose:
                logger.info(
                    f"Launched optimization for {plan_title} ({plan_id}): "
                    f"{optimization.operation_id}"
                )

    logger.info(
        "Finished initializing route optimizations for "
        f"{len(plan_df[~(plan_df[IntermediateColumns.ROUTE_TITLE]).isin(errors.keys())])} "
        "plans."
    )

    plan_df[IntermediateColumns.OPTIMIZED] = False
    plan_df.loc[
        (plan_df[IntermediateColumns.PLAN_ID].isin(plan_ids))
        & ~(plan_df[IntermediateColumns.ROUTE_TITLE].isin(errors.keys())),
        IntermediateColumns.OPTIMIZED,
    ] = True
    if errors:
        logger.warning(f"Errors launching optimizations:\n{errors}")

    plan_df = _confirm_optimizations(
        plan_df=plan_df, optimizations=optimizations, verbose=verbose
    )

    return plan_df


# TODO: Make a CLI for this since it will be optional.
# https://github.com/crickets-and-comb/bfb_delivery/issues/63
@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _distribute_routes(
    plan_df: DataFrame[schema.PlansDistributeRoutesIn], verbose: bool
) -> DataFrame[schema.PlansDistributeRoutesOut]:
    """Distribute the routes.

    Will skip plans marked as not initialized.

    Args:
        plan_df: The DataFrame with the plan IDs and driver IDs for each sheet.
        verbose: Whether to print verbose output.

    Returns:
        A DataFrame of the plans with a column for whether the routes were distributed.
    """
    logger.info("Distributing routes ...")

    plan_ids = plan_df[plan_df[IntermediateColumns.OPTIMIZED] == True][  # noqa: E712
        IntermediateColumns.PLAN_ID
    ].to_list()
    errors = {}
    for plan_id in plan_ids:
        plan_title = plan_df[plan_df[IntermediateColumns.PLAN_ID] == plan_id][
            IntermediateColumns.ROUTE_TITLE
        ].values[0]
        if verbose:
            logger.info(f"Distributing plan for {plan_title} ({plan_id}) ...")
        distributor = PlanDistributor(plan_id=plan_id, plan_title=plan_title)
        try:
            distributor.call_api()
        except Exception as e:
            logger.error(f"Error distributing plan for {plan_title} ({plan_id}):\n{e}")
            errors[plan_title] = e

    logger.info(
        "Finished distributing routes for "
        f"{len(plan_df[~(plan_df[IntermediateColumns.ROUTE_TITLE]).isin(errors.keys())])} "
        "plans."
    )
    plan_df[CircuitColumns.DISTRIBUTED] = False
    plan_df.loc[
        (plan_df[IntermediateColumns.PLAN_ID].isin(plan_ids))
        & ~(plan_df[IntermediateColumns.ROUTE_TITLE].isin(errors.keys())),
        CircuitColumns.DISTRIBUTED,
    ] = True
    if errors:
        logger.warning(f"Errors distributing plans:\n{errors}")

    return plan_df


@typechecked
def _print_report(plan_df: pd.DataFrame, no_distribute: bool) -> None:
    """Print a report of upload results."""
    plans_attempted = plan_df[IntermediateColumns.ROUTE_TITLE].to_list()
    plans_initialized = plan_df[
        plan_df[IntermediateColumns.INITIALIZED] == True  # noqa: E712
    ][IntermediateColumns.ROUTE_TITLE].to_list()
    plans_with_stops = plan_df[
        plan_df[IntermediateColumns.STOPS_UPLOADED] == True  # noqa: E712
    ][IntermediateColumns.ROUTE_TITLE].to_list()
    plans_optimized = plan_df[plan_df[IntermediateColumns.OPTIMIZED] == True][  # noqa: E712
        IntermediateColumns.ROUTE_TITLE
    ].to_list()
    plans_distributed = plan_df[plan_df[CircuitColumns.DISTRIBUTED] == True][  # noqa: E712
        IntermediateColumns.ROUTE_TITLE
    ].to_list()
    report_df = plan_df[
        [
            IntermediateColumns.ROUTE_TITLE,
            IntermediateColumns.INITIALIZED,
            CircuitColumns.WRITABLE,
            IntermediateColumns.STOPS_UPLOADED,
            IntermediateColumns.OPTIMIZED,
            CircuitColumns.DISTRIBUTED,
        ]
    ]
    logger.info(
        f"\n{report_df}\n"
        f"\nPlans attempted: {len(plans_attempted)}\n"
        f"Plans initialized: {len(plans_initialized)}\n"
        f"Plans with stops: {len(plans_with_stops)}\n"
        f"Plans optimized: {len(plans_optimized)}\n"
        f"Plans distributed: {len(plans_distributed)}"
    )
    if not no_distribute and len(plans_attempted) != len(plans_distributed):
        plans_not_distributed = list(set(plans_attempted) - set(plans_distributed))
        logger.warning(
            "Not all plans were distributed. "
            f"These are the undistributed plans:\n{plans_not_distributed}\n"
            "See the above output to see at which steps the plans failed."
        )
    elif no_distribute and len(plans_attempted) != len(plans_optimized):
        plans_not_optimized = list(set(plans_attempted) - set(plans_optimized))
        logger.warning(
            "Not all plans were optimized. "
            f"These are the non-optimized plans:\n{plans_not_optimized}\n"
            "See the above output to see at which steps the plans failed."
        )


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _assign_drivers_to_plans(
    stops_df: DataFrame[schema.Stops],
) -> DataFrame[schema.PlansAssignDriversToPlansOut]:
    """Get the driver IDs for each sheet.

    Args:
        stops_df: The long DataFrame with all the routes.

    Returns:
        A DataFrame with the driver IDs for each sheet.
    """
    plan_df = pd.DataFrame(
        {
            IntermediateColumns.ROUTE_TITLE: (
                stops_df[IntermediateColumns.SHEET_NAME].unique()
            ),
            IntermediateColumns.DRIVER_NAME: None,
            CircuitColumns.EMAIL: None,
            CircuitColumns.ID: None,
        }
    )

    drivers_df = _get_all_drivers()
    plan_df = _assign_drivers(drivers_df=drivers_df, plan_df=plan_df)
    inactive_drivers = plan_df[~(plan_df[CircuitColumns.ACTIVE])][
        IntermediateColumns.DRIVER_NAME
    ].tolist()
    if inactive_drivers:
        raise ValueError(
            (
                "Inactive drivers. Please activate the following drivers before creating "
                f"routes for them: {inactive_drivers}"
            )
        )

    return plan_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _initialize_plans(
    plan_df: DataFrame[schema.PlansInitializePlansIn], start_date: str, verbose: bool
) -> DataFrame[schema.PlansInitializePlansOut]:
    """Initialize Circuit plans with drivers."""
    plan_df[IntermediateColumns.PLAN_ID] = None
    plan_df[CircuitColumns.WRITABLE] = None
    # TODO: Do we need this column?
    # https://github.com/crickets-and-comb/bfb_delivery/issues/64
    plan_df[CircuitColumns.OPTIMIZATION] = None

    logger.info("Initializing plans ...")
    errors = {}
    for idx, row in plan_df.iterrows():
        plan_data = {
            CircuitColumns.TITLE: row[IntermediateColumns.ROUTE_TITLE],
            # TODO: Just make this once.
            # https://github.com/crickets-and-comb/bfb_delivery/issues/65
            CircuitColumns.STARTS: {
                CircuitColumns.DAY: int(start_date.split("-")[2]),
                CircuitColumns.MONTH: int(start_date.split("-")[1]),
                CircuitColumns.YEAR: int(start_date.split("-")[0]),
            },
            CircuitColumns.DRIVERS: [row[CircuitColumns.ID]],
        }
        if verbose:
            logger.info(f"Creating plan for {row[IntermediateColumns.ROUTE_TITLE]} ...")
        plan_initializer = PlanInitializer(plan_data=plan_data)
        try:
            plan_initializer.call_api()
        except Exception as e:
            logger.error(
                f"Error initializing plan for {row[IntermediateColumns.ROUTE_TITLE]}:\n{e}"
            )
            errors[row[IntermediateColumns.ROUTE_TITLE]] = e
        else:
            if verbose:
                logger.info(
                    f"Created plan {plan_initializer.response_json[CircuitColumns.ID]} for "
                    f"{row[IntermediateColumns.ROUTE_TITLE]}."
                    f"\n{plan_initializer.response_json}"
                )
        finally:
            plan_df.loc[idx, IntermediateColumns.PLAN_ID] = (
                plan_initializer.plan_id
                if "plan_id" in vars(plan_initializer)
                else "plans/noID"
            )
            plan_df.loc[idx, CircuitColumns.WRITABLE] = (
                plan_initializer.writable if "writable" in vars(plan_initializer) else False
            )

    logger.info(f"Finished initializing plans. Initialized {idx + 1 - len(errors)} plans.")

    plan_df[IntermediateColumns.INITIALIZED] = True
    plan_df.loc[
        (plan_df[IntermediateColumns.ROUTE_TITLE].isin(errors.keys()))
        # TODO: Make not_writable a failure within class?
        # https://github.com/crickets-and-comb/bfb_delivery/issues/66
        | ~(plan_df[CircuitColumns.WRITABLE] == True),  # noqa: E712
        IntermediateColumns.INITIALIZED,
    ] = False

    if errors:
        logger.warning(f"Errors initializing plans:\n{errors}")

    not_writable = plan_df[~(plan_df[CircuitColumns.WRITABLE] == True)]  # noqa: E712
    # TODO: Add noqa E712 to shared, and remove throughout codebase.
    # https://github.com/crickets-and-comb/shared/issues/41
    if not not_writable.empty:
        logger.warning(f"Plan is not writable for the following routes:\n{not_writable}")

    return plan_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _build_plan_stops(
    stops_df: DataFrame[schema.Stops], plan_df: DataFrame[schema.PlansBuildStopsIn]
) -> dict[str, list[list[dict[str, dict[str, str] | list[str] | int | str]]]]:
    """Build stop arrays for each route.

    Args:
        stops_df: The long DataFrame with all the routes.
        plan_df: The DataFrame with the plan IDs and driver IDs for each sheet.

    Returns:
        For each plan, a list of stop arrays for batch stop uploads.
    """
    stops_df = _parse_addresses(stops_df=stops_df)
    plan_stops = {}
    for _, plan_row in plan_df.iterrows():
        plan_id = plan_row[IntermediateColumns.PLAN_ID]
        route_title = plan_row[IntermediateColumns.ROUTE_TITLE]
        route_stops = stops_df[stops_df[IntermediateColumns.SHEET_NAME] == route_title]
        plan_stops[plan_id] = _build_stop_array(
            route_stops=route_stops, driver_id=plan_row[CircuitColumns.ID]
        )

    for plan_id, all_stops in plan_stops.items():
        stop_arrays = []
        # Split all_stops into chunks of 100 stops.
        number_of_stops = len(all_stops)
        for i in range(0, number_of_stops, RateLimits.BATCH_STOP_IMPORT_MAX_STOPS):
            stop_arrays.append(
                all_stops[i : i + 100]  # noqa: E203
            )  # TODO: Add noqa E203 to shared, and remove throughout codebase.
            # https://github.com/crickets-and-comb/shared/issues/41
        plan_stops[plan_id] = stop_arrays

    return plan_stops


# TODO: Why isn't this throwing when driver ID is invalid?
# https://github.com/crickets-and-comb/bfb_delivery/issues/67
@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _get_all_drivers() -> schema.DriversGetAllDriversOut:
    """Get all drivers."""
    logger.info("Getting all drivers from Circuit ...")
    driver_pages = get_responses(
        url=CIRCUIT_DRIVERS_URL, paged_response_class=PagedResponseGetterBFB
    )
    logger.info("Finished getting drivers.")
    drivers_list = concat_response_pages(
        page_list=driver_pages, data_key=CircuitColumns.DRIVERS
    )
    drivers_df = pd.DataFrame(drivers_list)
    drivers_df = drivers_df.sort_values(by=CircuitColumns.NAME).reset_index(drop=True)

    return drivers_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _assign_drivers(
    drivers_df: DataFrame[schema.DriversAssignDriversIn],
    plan_df: DataFrame[schema.PlansAssignDriversIn],
) -> DataFrame[schema.PlansAssignDriversOut]:
    """Ask users to assign drivers to each route."""
    for idx, row in drivers_df.iterrows():
        print(
            f"{idx + 1}. {'Active' if row[CircuitColumns.ACTIVE] else 'Inactive'}: "
            f"{row[CircuitColumns.NAME]} {row[CircuitColumns.EMAIL]}"
        )

    print("\nUsing the driver numbers above, assign drivers to each route:")
    for route_title in plan_df[IntermediateColumns.ROUTE_TITLE]:
        plan_df = _assign_driver(
            route_title=route_title, drivers_df=drivers_df, plan_df=plan_df
        )

    for _, row in plan_df.iterrows():
        print(
            f"{row[IntermediateColumns.ROUTE_TITLE]}: "
            f"{row[IntermediateColumns.DRIVER_NAME]}, {row[CircuitColumns.EMAIL]}"
        )
    confirm = input("Confirm the drivers above? (y/n): ")
    # TODO: Check for y, n, and prompt again if neither.
    # https://github.com/crickets-and-comb/bfb_delivery/issues/68
    if confirm.lower() != "y":
        plan_df = _assign_drivers(drivers_df=drivers_df, plan_df=plan_df)

    return plan_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
# TODO: Decrease complexity. (Remove noqa C901.)
# https://github.com/crickets-and-comb/bfb_delivery/issues/69
def _assign_driver(  # noqa: C901
    route_title: str,
    drivers_df: DataFrame[schema.DriversAssignDriverIn],
    plan_df: DataFrame[schema.PlansAssignDriverIn],
) -> pd.DataFrame:
    """Ask user to assign driver to a route."""
    best_guesses = pd.DataFrame()
    for name_part in route_title.split(" ")[1:]:
        if name_part not in ["&", "AND"] and len(name_part) > 1:
            best_guesses = pd.concat(
                [
                    best_guesses,
                    drivers_df[
                        drivers_df[CircuitColumns.NAME].str.contains(name_part, case=False)
                    ],
                ]
            )
    # Using ID with name/email as added validation of our assumptions about uniqueness.
    # Should break more loudly if so than if we only used ID or name/email compound key.
    id_cols = [CircuitColumns.ID, CircuitColumns.NAME, CircuitColumns.EMAIL]
    best_guesses = best_guesses.drop_duplicates(subset=id_cols).sort_values(
        by=CircuitColumns.NAME
    )

    print(f"\nRoute {route_title}:\nBest guesses:")
    for idx, driver in best_guesses.iterrows():
        print(
            f"{idx + 1}. {'Active' if driver[CircuitColumns.ACTIVE] else 'Inactive'}: "
            f"{driver[CircuitColumns.NAME]} {driver[CircuitColumns.EMAIL]}"
        )
    print("\n")

    assigned = False
    while not assigned:
        try:
            # TODO: Add B907 to shared ignore list, and remove r"" throughout.
            # https://github.com/crickets-and-comb/shared/issues/41
            # TODO: Add option to correct the previous.
            # https://github.com/crickets-and-comb/bfb_delivery/issues/70
            choice = input(
                f"Enter the number of the driver for '{route_title}'"  # noqa: B907
                "(ctl+c to start over):"
            )
        except ValueError:
            print("Invalid input. Please enter a number.")

        else:
            choice = choice if choice else "-1"
            try:
                choice = int(choice.strip()) - 1
                if choice < 0 or choice >= len(drivers_df):
                    raise errors.AssignmentOutOfRange
                driver = drivers_df.iloc[choice]
                if not driver[CircuitColumns.ACTIVE]:
                    raise errors.InactiveDriverAssignment
            except ValueError:
                print("Invalid input. Please enter a number.")
            except errors.AssignmentOutOfRange:
                print("Invalid input. Please enter a number associated with a driver.")
            except errors.InactiveDriverAssignment:
                print("Inactive driver selected. Select an active driver.")
            else:
                plan_df.loc[
                    plan_df[IntermediateColumns.ROUTE_TITLE] == route_title,
                    [
                        CircuitColumns.ID,
                        IntermediateColumns.DRIVER_NAME,
                        CircuitColumns.EMAIL,
                        CircuitColumns.ACTIVE,
                    ],
                ] = [
                    drivers_df.iloc[choice][CircuitColumns.ID],
                    drivers_df.iloc[choice][CircuitColumns.NAME],
                    drivers_df.iloc[choice][CircuitColumns.EMAIL],
                    drivers_df.iloc[choice][CircuitColumns.ACTIVE],
                ]
                assigned = True
                print(
                    f"\nAssigned {route_title} "
                    f"to {drivers_df.iloc[choice][CircuitColumns.NAME]}."
                )

    return plan_df


@typechecked
def _parse_addresses(stops_df: pd.DataFrame) -> pd.DataFrame:
    """Parse addresses for each route."""
    stops_df[CircuitColumns.ADDRESS_NAME] = stops_df[Columns.ADDRESS]
    stops_df[CircuitColumns.ADDRESS_LINE_1] = ""
    stops_df[CircuitColumns.ADDRESS_LINE_2] = ""
    stops_df[CircuitColumns.STATE] = "WA"
    stops_df[CircuitColumns.ZIP] = ""
    stops_df[CircuitColumns.COUNTRY] = "US"

    for idx, row in stops_df.iterrows():
        address = row[Columns.ADDRESS]
        split_address = address.split(",")
        stops_df.at[idx, CircuitColumns.ADDRESS_LINE_1] = split_address[0].strip()
        stops_df.at[idx, CircuitColumns.ADDRESS_LINE_2] = ", ".join(
            [part.strip() for part in split_address[1:]]
        )
        stops_df.at[idx, CircuitColumns.ZIP] = split_address[-1]

    return stops_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _confirm_optimizations(
    plan_df: DataFrame[schema.PlansConfirmOptimizationsIn],
    optimizations: dict[str, str],
    verbose: bool,
) -> DataFrame[schema.PlansConfirmOptimizationsOut]:
    """Confirm all optimizations have finished."""
    logger.info("Confirming optimizations have finished ...")

    optimizations_finished: dict[str, bool | str] = {
        plan_id: False
        for plan_id in plan_df[plan_df[IntermediateColumns.OPTIMIZED] == True][  # noqa: E712
            IntermediateColumns.PLAN_ID
        ].to_list()
    }
    errors = {}

    while not all([val is True or val == "error" for val in optimizations_finished.values()]):
        unfinished = [
            plan_id
            for plan_id, finished in optimizations_finished.items()
            if not finished or finished != "error"
        ]
        for plan_id in unfinished:
            plan_title = plan_df[plan_df[IntermediateColumns.PLAN_ID] == plan_id][
                IntermediateColumns.ROUTE_TITLE
            ].values[0]
            if verbose:
                logger.info(f"Checking optimization for {plan_title} ({plan_id}) ...")
            check_op = OptimizationChecker(
                plan_id=plan_id, operation_id=optimizations[plan_id], plan_title=plan_title
            )
            try:
                check_op.call_api()
            except Exception as e:
                logger.error(
                    f"Error checking optimization for {plan_title} ({plan_id}):\n{e}"
                )
                errors[plan_title] = [e]

                optimizations_finished[plan_id] = "error"

            else:
                optimizations_finished[plan_id] = check_op.finished
                if verbose:
                    logger.info(
                        f"Optimization status for {plan_title} ({plan_id}): "
                        f"{check_op.finished}"
                    )

    logger.info(
        "Finished optimizing routes. Optimized "
        f"{len([val for val in optimizations_finished.values() if val is True])} routes."
    )
    # TODO: FutureWarning: Setting an item of incompatible dtype is deprecated and will raise
    # an error in a future version of pandas. Value 'nan' has dtype incompatible with bool,
    # please explicitly cast to a compatible dtype first.
    # https://github.com/crickets-and-comb/bfb_delivery/issues/71
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)
        plan_df.loc[
            plan_df[IntermediateColumns.ROUTE_TITLE].isin(errors.keys()),
            IntermediateColumns.OPTIMIZED,
        ] = None
    if errors:
        logger.warning(f"Errors checking optimizations:\n{errors}")

    return plan_df


@typechecked
def _build_stop_array(route_stops: pd.DataFrame, driver_id: str) -> list[dict[str, Any]]:
    """Build a stop array for a route."""
    stop_array = []
    for _, stop_row in route_stops.iterrows():
        stop = {
            CircuitColumns.ADDRESS: {
                CircuitColumns.ADDRESS_NAME: stop_row[CircuitColumns.ADDRESS_NAME],
                CircuitColumns.ADDRESS_LINE_1: stop_row[CircuitColumns.ADDRESS_LINE_1],
                CircuitColumns.ADDRESS_LINE_2: stop_row[CircuitColumns.ADDRESS_LINE_2],
                CircuitColumns.STATE: stop_row[CircuitColumns.STATE],
                CircuitColumns.ZIP: stop_row[CircuitColumns.ZIP],
                CircuitColumns.COUNTRY: stop_row[CircuitColumns.COUNTRY],
            },
            CircuitColumns.ORDER_INFO: {
                CircuitColumns.PRODUCTS: [stop_row[Columns.PRODUCT_TYPE]]
            },
            CircuitColumns.ALLOWED_DRIVERS: [driver_id],
            CircuitColumns.PACKAGE_COUNT: stop_row[Columns.ORDER_COUNT],
        }

        if stop_row.get(Columns.NOTES) and not pd.isna(stop_row[Columns.NOTES]):
            stop[CircuitColumns.NOTES] = stop_row[Columns.NOTES]

        recipient_dict = {}
        if stop_row.get(Columns.EMAIL) and not pd.isna(stop_row[Columns.EMAIL]):
            recipient_dict[CircuitColumns.EMAIL] = stop_row[Columns.EMAIL]
        if stop_row.get(Columns.PHONE) and not pd.isna(stop_row[Columns.PHONE]):
            recipient_dict[CircuitColumns.PHONE] = stop_row[Columns.PHONE]
        if stop_row.get(Columns.NAME) and not pd.isna(stop_row[Columns.NAME]):
            recipient_dict[CircuitColumns.NAME] = stop_row[Columns.NAME]
        if stop_row.get(Columns.NEIGHBORHOOD) and not pd.isna(stop_row[Columns.NEIGHBORHOOD]):
            recipient_dict[CircuitColumns.EXTERNAL_ID] = stop_row[Columns.NEIGHBORHOOD]
        if recipient_dict:
            stop[CircuitColumns.RECIPIENT] = recipient_dict

        stop_array.append(stop)

    return stop_array



"""Read from Circuit."""

import logging
import os
import shutil
from pathlib import Path
from typing import Any

import pandas as pd
import pandera as pa
from pandera.typing import DataFrame
from typeguard import typechecked

from comb_utils import concat_response_pages, get_responses

from bfb_delivery.lib.constants import (
    ALL_HHS_DRIVER,
    CIRCUIT_DOWNLOAD_COLUMNS,
    DEPOT_PLACE_ID,
    CircuitColumns,
    Columns,
    IntermediateColumns,
)
from bfb_delivery.lib.dispatch.api_callers import PagedResponseGetterBFB
from bfb_delivery.lib.schema import (
    CircuitPlansFromDict,
    CircuitPlansOut,
    CircuitPlansTransformIn,
    CircuitRoutesTransformInFromDict,
    CircuitRoutesTransformOut,
    CircuitRoutesWriteIn,
    CircuitRoutesWriteOut,
)
from bfb_delivery.lib.schema.utils import schema_error_handler
from bfb_delivery.lib.utils import get_friday

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@typechecked
def get_route_files(
    start_date: str,
    end_date: str,
    plan_ids: list[str],
    output_dir: str,
    all_hhs: bool = False,
    verbose: bool = False,
) -> str:
    """Get the route files for the given date.

    Args:
        start_date: The start date to get the routes for, as "YYYYMMDD".
            Empty string uses the soonest Friday.
        end_date: The end date to get the routes for, as "YYYYMMDD".
            Empty string uses the start date.
        plan_ids: The plan IDs to get the routes for. Overrides `all_hhs`.
        output_dir: The directory to create a subdir to save the routes to.
            Creates "routes_{date}" directory within the `output_dir`.
            Empty string uses the current working directory.
            If the directory does not exist, it is created. If it exists, it is overwritten.
        all_hhs: Flag to get only the "All HHs" route.
            False gets all routes except "All HHs". True gets only the "All HHs" route.
            Overriden by `plan_ids`.
        verbose: Flag to print verbose output.

    Returns:
        The path to the route files.

    Raises:
        ValueError: If no plans are found for the given date range.
        ValueError: If no plans with routes are found.
        ValueError: If no stops are found for the given plans.
        ValueError: If no routed stops are found for the given plans.
    """
    start_date = start_date if start_date else get_friday(fmt="%Y%m%d")
    end_date = end_date if end_date else start_date
    sub_dir = "routes_" + start_date
    output_dir = (
        str(Path(output_dir) / sub_dir) if output_dir else str(Path(_getcwd()) / sub_dir)
    )

    plans_list = _get_raw_plans(start_date=start_date, end_date=end_date, verbose=verbose)
    plans_df = _make_plans_df(
        plans_list=plans_list, plan_ids=plan_ids, all_hhs=all_hhs, verbose=verbose
    )
    plan_stops_list = _get_raw_stops(
        plan_ids=plans_df[CircuitColumns.ID].tolist(), verbose=verbose
    )
    routes_df = _transform_routes_df(
        plan_stops_list=plan_stops_list, plans_df=plans_df, verbose=verbose
    )
    _write_routes_dfs(routes_df=routes_df, output_dir=Path(output_dir))

    return output_dir


def _getcwd() -> str:
    """Wrapping to allow restricted mocking."""
    return os.getcwd()


@typechecked
def _get_raw_plans(start_date: str, end_date: str, verbose: bool) -> list[dict[str, Any]]:
    """Call Circuit API to get the plans for the given date."""
    url = (
        "https://api.getcircuit.com/public/v0.2b/plans"
        f"?filter.startsGte={start_date}"
        f"&filter.startsLte={end_date}"
    )
    logger.info("Getting route plans from Circuit ...")
    if verbose:
        logger.info(f"Getting route plans from {url} ...")
    plans = _get_plan_responses(url=url)
    logger.info("Finished getting route plans.")
    plans_list = concat_response_pages(page_list=plans, data_key="plans")

    if not plans_list:
        raise ValueError(f"No plans found for {start_date} to {end_date}.")
    return plans_list


# Using from_format config https://pandera.readthedocs.io/en/v0.22.1/
# data_format_conversion.html#data-format-conversion
# Here, you pass in plans_df as a list of dictionaries, but you treat/type it as a dataframe.
# Not a huge fan as it obscures the pipeline steps and makes it a little harder to follow.
# But, it makes input validation simpler.
# NOTE: with_pydantic allows check of other params.
@pa.check_types(with_pydantic=True, lazy=True)
def _make_plans_df(
    plans_list: DataFrame[CircuitPlansFromDict],
    all_hhs: bool,
    plan_ids: list[str] | None = None,
    verbose: bool = False,
) -> DataFrame[CircuitPlansOut]:
    """Make the plans DataFrame from the plans."""
    # What we'd do if not using from_format config:
    # plans_df = pd.DataFrame(plans_list)

    plan_count = len(plans_list)
    plan_mask = [True] * plan_count
    if not plan_ids:
        if all_hhs:
            plan_mask = [
                ALL_HHS_DRIVER.upper() in title.upper()
                for title in plans_list[CircuitColumns.TITLE]
            ]
        else:
            plan_mask = [
                ALL_HHS_DRIVER.upper() not in title.upper()
                for title in plans_list[CircuitColumns.TITLE]
            ]

        if verbose:
            _count_allhhs_dropped(all_hhs=all_hhs, plan_count=plan_count, plan_mask=plan_mask)

    else:
        plan_mask = [plan_id in plan_ids for plan_id in plans_list[CircuitColumns.ID]]
        if verbose:
            _count_plan_ids_dropped(plan_count=plan_count, plan_mask=plan_mask)

    plans_df = plans_list[plan_mask]
    plan_count = len(plans_df)

    if all_hhs and plan_count != 1 and verbose:
        logger.warning(
            f'Got {plan_count} "{ALL_HHS_DRIVER}" plans. Expected 1.'  # noqa: B907
        )

    routed_plans_mask = [isinstance(val, list) and len(val) > 0 for val in plans_df["routes"]]
    plans_df = plans_df[routed_plans_mask]
    if verbose:
        logger.info("Filtering out plans with no routes.")
        dropped_count = plan_count - len(plans_df)
        if dropped_count:
            logger.warning(f"Dropped {dropped_count} plans without routes.")
        else:
            logger.info("Dropped no plans.")

    plan_count = len(plans_df)
    if not plan_count:
        raise ValueError("No routes found for the given date range.")
    elif verbose:
        logger.info(f"Left with {plan_count} plans.")

    plans_df = plans_df[[CircuitColumns.ID, CircuitColumns.TITLE]]

    return plans_df


@typechecked
def _count_allhhs_dropped(all_hhs: bool, plan_count: int, plan_mask: list[bool]) -> None:
    if all_hhs:
        logger.info(f'Filtered to only the "{ALL_HHS_DRIVER}" plan.')  # noqa: B907
    else:
        logger.info(f'Filtered to all plans except "{ALL_HHS_DRIVER}".')  # noqa: B907

    dropped_count = plan_count - sum(plan_mask)
    if not all_hhs and dropped_count != 1:
        logger.warning(f"Dropped {dropped_count} plans.")
    elif dropped_count:
        logger.info(f"Dropped {dropped_count} plans.")
    else:
        logger.info("Dropped no plans.")


@typechecked
def _count_plan_ids_dropped(plan_count: int, plan_mask: list[bool]) -> None:
    logger.info("Filtering to specified plan IDs.")
    dropped_count = plan_count - sum(plan_mask)
    if dropped_count:
        logger.warning(f"Dropped {dropped_count} plans.")
    else:
        logger.info("Dropped no plans.")


@typechecked
def _get_raw_stops(plan_ids: list[str], verbose: bool) -> list[dict[str, Any]]:
    """Get the raw stops list from Circuit."""
    logger.info("Getting stops from Circuit ...")
    stops_lists_list = _get_raw_stops_list(plan_ids=plan_ids, verbose=verbose)
    logger.info("Finished getting stops.")

    plan_stops_list = []
    for stop_lists in stops_lists_list:
        plan_stops_list += concat_response_pages(
            page_list=stop_lists, data_key=CircuitColumns.STOPS
        )
    if not plan_stops_list:
        raise ValueError(f"No stops found for plans {plan_ids}.")

    return plan_stops_list


@typechecked
def _get_raw_stops_list(plan_ids: list[str], verbose: bool) -> list[Any]:
    stops_lists_list = []
    for plan_id in plan_ids:
        # https://developer.team.getcircuit.com/api#tag/Stops/operation/listStops
        url = f"https://api.getcircuit.com/public/v0.2b/{plan_id}/stops"
        if verbose:
            logger.info(f"Getting stops from {url} ...")
        stops_lists = _get_stops_responses(url=url)
        stops_lists_list.append(stops_lists)

    return stops_lists_list


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _transform_routes_df(
    plan_stops_list: DataFrame[CircuitRoutesTransformInFromDict],
    plans_df: DataFrame[CircuitPlansTransformIn],
    verbose: bool = False,
) -> DataFrame[CircuitRoutesTransformOut]:
    """Transform the raw routes DataFrame."""
    routes_df = _pare_routes_df(routes_df=plan_stops_list, verbose=verbose)
    del plan_stops_list

    routes_df = routes_df.merge(
        plans_df.copy().rename(columns={CircuitColumns.ID: "plan_id"}),
        left_on=CircuitColumns.PLAN,
        right_on="plan_id",
        how="left",
        validate="m:1",
    )

    routes_df.rename(
        columns={
            # Plan title is upload/download sheet name.
            CircuitColumns.TITLE: IntermediateColumns.DRIVER_SHEET_NAME,
            CircuitColumns.STOP_POSITION: Columns.STOP_NO,
            CircuitColumns.NOTES: Columns.NOTES,
            CircuitColumns.PACKAGE_COUNT: Columns.ORDER_COUNT,
            CircuitColumns.ADDRESS: Columns.ADDRESS,
        },
        inplace=True,
    )

    routes_df = _set_routes_df_values(routes_df=routes_df)

    routes_df.sort_values(
        by=[IntermediateColumns.DRIVER_SHEET_NAME, Columns.STOP_NO], inplace=True
    )

    return routes_df


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _write_routes_dfs(routes_df: DataFrame[CircuitRoutesWriteIn], output_dir: Path) -> None:
    """Split and write the routes DataFrame to the output directory.

    Args:
        routes_df: The routes DataFrame to write.
        output_dir: The directory to save the routes to.
    """
    if output_dir.exists():
        logger.warning(f"Output directory exists {output_dir}. Overwriting.")
        shutil.rmtree(output_dir, ignore_errors=True)
    output_dir.mkdir(parents=True)

    logger.info(f"Writing route CSVs to {output_dir.resolve()}")
    for route, route_df in routes_df.groupby(CircuitColumns.ROUTE):
        driver_sheet_names = route_df[IntermediateColumns.DRIVER_SHEET_NAME].unique()
        if len(driver_sheet_names) > 1:
            raise ValueError(
                f"Route {route} has multiple driver sheet names: {driver_sheet_names}"
            )
        elif len(driver_sheet_names) < 1:
            raise ValueError(f"Route {route} has no driver sheet name.")

        route_df = route_df[CIRCUIT_DOWNLOAD_COLUMNS]
        driver_sheet_name = driver_sheet_names[0]
        fp = output_dir / f"{driver_sheet_name}.csv"
        _write_route_df(route_df=route_df, fp=fp)


@schema_error_handler
@pa.check_types(with_pydantic=True, lazy=True)
def _write_route_df(route_df: DataFrame[CircuitRoutesWriteOut], fp: Path) -> None:
    route_df.to_csv(fp, index=False)


@typechecked
def _get_plan_responses(url: str) -> list[dict[str, Any]]:
    return get_responses(url=url, paged_response_class=PagedResponseGetterBFB)


@typechecked
def _get_stops_responses(url: str) -> list[dict[str, Any]]:
    return get_responses(url=url, paged_response_class=PagedResponseGetterBFB)


@typechecked
def _pare_routes_df(routes_df: pd.DataFrame, verbose: bool) -> pd.DataFrame:
    routes_df = routes_df[
        [
            # plan id e.g. "plans/0IWNayD8NEkvD5fQe2SQ":
            CircuitColumns.PLAN,
            CircuitColumns.ROUTE,
            # stop id e.g. "plans/0IWNayD8NEkvD5fQe2SQ/stops/40lmbcQrd32NOfZiiC1b":
            CircuitColumns.ID,
            CircuitColumns.STOP_POSITION,
            CircuitColumns.RECIPIENT,
            CircuitColumns.ADDRESS,
            CircuitColumns.NOTES,
            CircuitColumns.ORDER_INFO,
            CircuitColumns.PACKAGE_COUNT,
        ]
    ]

    if verbose:
        logger.info("Filtering out stops without routes.")
        stop_count = len(routes_df)

    routed_stops_mask = [
        isinstance(route_dict, dict) and route_dict.get(CircuitColumns.ID, "") != ""
        for route_dict in routes_df[CircuitColumns.ROUTE]
    ]
    routes_df = routes_df[routed_stops_mask]
    if verbose:
        dropped_count = stop_count - len(routes_df)
        if dropped_count:
            logger.warning(f"Dropped {dropped_count} stops without routes.")
        else:
            logger.info("Dropped no stops.")

        logger.info("Filtering out depot stops.")
        stop_count = len(routes_df)

    routes_df[CircuitColumns.PLACE_ID] = routes_df[CircuitColumns.ADDRESS].apply(
        lambda address_dict: address_dict.get(CircuitColumns.PLACE_ID)
    )
    routes_df = routes_df[routes_df[CircuitColumns.PLACE_ID] != DEPOT_PLACE_ID]

    if verbose:
        dropped_count = stop_count - len(routes_df)
        logger.info(f"Dropped {dropped_count} stops.")

    stop_count = len(routes_df)
    if not stop_count:
        raise ValueError("No routed stops found for the given plans.")
    elif verbose:
        logger.info(f"Left with {stop_count} routed stops.")

    return routes_df


@typechecked
def _set_routes_df_values(routes_df: pd.DataFrame) -> pd.DataFrame:
    routes_df[IntermediateColumns.ROUTE_TITLE] = routes_df[CircuitColumns.ROUTE].apply(
        lambda route_dict: route_dict.get(CircuitColumns.TITLE)
    )
    routes_df[CircuitColumns.ROUTE] = routes_df[CircuitColumns.ROUTE].apply(
        lambda route_dict: route_dict.get(CircuitColumns.ID)
    )
    routes_df[Columns.NAME] = routes_df[CircuitColumns.RECIPIENT].apply(
        lambda recipient_dict: recipient_dict.get(CircuitColumns.NAME)
    )
    routes_df[CircuitColumns.ADDRESS_LINE_1] = routes_df[Columns.ADDRESS].apply(
        lambda address_dict: address_dict.get(CircuitColumns.ADDRESS_LINE_1)
    )
    routes_df[CircuitColumns.ADDRESS_LINE_2] = routes_df[Columns.ADDRESS].apply(
        lambda address_dict: address_dict.get(CircuitColumns.ADDRESS_LINE_2)
    )
    routes_df[Columns.PHONE] = routes_df[CircuitColumns.RECIPIENT].apply(
        lambda recipient_dict: recipient_dict.get(CircuitColumns.PHONE)
    )
    routes_df[Columns.BOX_TYPE] = routes_df[CircuitColumns.ORDER_INFO].apply(
        lambda order_info_dict: (
            order_info_dict[CircuitColumns.PRODUCTS][0]
            if order_info_dict.get(CircuitColumns.PRODUCTS)
            else None
        )
    )
    routes_df[Columns.NEIGHBORHOOD] = routes_df[CircuitColumns.RECIPIENT].apply(
        lambda recipient_dict: recipient_dict.get(CircuitColumns.EXTERNAL_ID)
    )
    routes_df[Columns.EMAIL] = routes_df[CircuitColumns.RECIPIENT].apply(
        lambda recipient_dict: recipient_dict.get(CircuitColumns.EMAIL)
    )
    routes_df[IntermediateColumns.DRIVER_SHEET_NAME] = _clean_title(
        routes_df[IntermediateColumns.DRIVER_SHEET_NAME], warn=True
    )
    routes_df[IntermediateColumns.ROUTE_TITLE] = _clean_title(
        routes_df[IntermediateColumns.ROUTE_TITLE], warn=False
    )

    _warn_and_impute(routes_df=routes_df)

    routes_df[Columns.ADDRESS] = (
        routes_df[CircuitColumns.ADDRESS_LINE_1]
        + ", "
        + routes_df[CircuitColumns.ADDRESS_LINE_2]
    )

    _split_multi_route_drivers(routes_df=routes_df)

    return routes_df


@typechecked
def _clean_title(title_series: pd.Series, warn: bool) -> pd.Series:
    """Clean the title column."""
    for title in title_series.unique():
        if "/" in title and warn:
            logger.warning(f'Title "{title}" contains "/". Replacing with ".".')  # noqa: B907
    return title_series.str.replace("/", ".")


@typechecked
def _warn_and_impute(routes_df: pd.DataFrame) -> None:
    """Warn and impute missing values in the routes DataFrame."""
    missing_order_count = routes_df[Columns.ORDER_COUNT].isna()
    if missing_order_count.any():
        logger.warning(
            f"Missing order count for {missing_order_count.sum()} stops. Imputing 1 order."
        )
    routes_df[Columns.ORDER_COUNT] = routes_df[Columns.ORDER_COUNT].fillna(1)

    missing_neighborhood = routes_df[Columns.NEIGHBORHOOD].isna()
    if missing_neighborhood.any():
        logger.warning(
            f"Missing neighborhood for {missing_neighborhood.sum()} stops. "
            "Imputing best guesses from Circuit-supplied address."
        )
    routes_df[Columns.NEIGHBORHOOD] = routes_df[
        [Columns.NEIGHBORHOOD, Columns.ADDRESS]
    ].apply(
        lambda row: (
            row[Columns.NEIGHBORHOOD]
            if row[Columns.NEIGHBORHOOD]
            else row[Columns.ADDRESS].get(CircuitColumns.ADDRESS).split(",")[1].strip()
        ),
        axis=1,
    )


@typechecked
def _split_multi_route_drivers(routes_df: pd.DataFrame) -> None:
    """If a driver sheet name has multiple routes, split them into separate sheet names."""
    many_routes_to_one_driver = routes_df.groupby(IntermediateColumns.DRIVER_SHEET_NAME)[
        CircuitColumns.ROUTE
    ].nunique()
    many_routes_to_one_driver = many_routes_to_one_driver[
        many_routes_to_one_driver > 1
    ].index.tolist()
    for sheet_name in many_routes_to_one_driver:
        routes = routes_df[routes_df[IntermediateColumns.DRIVER_SHEET_NAME] == sheet_name][
            CircuitColumns.ROUTE
        ].unique()
        route_count = len(routes)
        logger.warning(
            f'Manifest "{sheet_name}" has {route_count} routes. '  # noqa: B907
            f'Renaming them to "{sheet_name} #<1-{route_count}>"'
        )
        for i, route in enumerate(routes):
            for title_col in [
                IntermediateColumns.DRIVER_SHEET_NAME,
                IntermediateColumns.ROUTE_TITLE,
            ]:
                routes_df.loc[routes_df[CircuitColumns.ROUTE] == route, title_col] = (
                    f"{sheet_name} #{i + 1}"
                )



# noqa: D100
__doc__ = """
.. click:: bfb_delivery.cli.create_manifests:main
   :prog: create_manifests
   :nested: full
"""

import logging

import click
from typeguard import typechecked

from bfb_delivery import create_manifests
from bfb_delivery.lib.constants import Defaults, DocStrings

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@click.command(help=DocStrings.CREATE_MANIFESTS.cli_docstring)
@click.option(
    "--input_dir", type=str, required=True, help=DocStrings.CREATE_MANIFESTS.args["input_dir"]
)
@click.option(
    "--output_dir",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS["output_dir"],
    help=DocStrings.CREATE_MANIFESTS.args["output_dir"],
)
@click.option(
    "--output_filename",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS["output_filename"],
    help=DocStrings.CREATE_MANIFESTS.args["output_filename"],
)
@click.option(
    "--extra_notes_file",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS["extra_notes_file"],
    help=DocStrings.CREATE_MANIFESTS.args["extra_notes_file"],
)
@typechecked
def main(  # noqa: D103
    input_dir: str, output_dir: str, output_filename: str, extra_notes_file: str
) -> str:
    final_manifest_path = create_manifests(
        input_dir=input_dir,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )
    logger.info(f"Final manifests saved to:\n{final_manifest_path.resolve()}")

    return str(final_manifest_path)



# noqa: D100
__doc__ = """
.. click:: bfb_delivery.cli.combine_route_tables:main
   :prog: combine_route_tables
   :nested: full
"""
import logging

import click
from typeguard import typechecked

from bfb_delivery import combine_route_tables
from bfb_delivery.lib.constants import Defaults, DocStrings

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@click.command(help=DocStrings.COMBINE_ROUTE_TABLES.cli_docstring)
@click.option(
    "--input_dir",
    type=str,
    required=True,
    help=DocStrings.COMBINE_ROUTE_TABLES.args["input_dir"],
)
@click.option(
    "--output_dir",
    type=str,
    required=False,
    default=Defaults.COMBINE_ROUTE_TABLES["output_dir"],
    help=DocStrings.COMBINE_ROUTE_TABLES.args["output_dir"],
)
@click.option(
    "--output_filename",
    type=str,
    required=False,
    default=Defaults.COMBINE_ROUTE_TABLES["output_filename"],
    help=DocStrings.COMBINE_ROUTE_TABLES.args["output_filename"],
)
@typechecked
def main(input_dir: str, output_dir: str, output_filename: str) -> str:  # noqa: D103
    path = combine_route_tables(
        input_dir=input_dir, output_dir=output_dir, output_filename=output_filename
    )
    logger.info(f"Combined workbook saved to:\n{path.resolve()}")

    return str(path)



"""Constants used in the project."""

from enum import StrEnum
from typing import Final

import pandas as pd

from comb_utils import DocString, ErrorDocString

ADDRESS_COLUMN_WIDTH: Final[float] = 40

ALL_HHS_DRIVER: Final[str] = "All HHs"


class BookOneDrivers(StrEnum):
    """Drivers for the first book.

    This is only an enum so it appears in docs.
    """

    DUMMY = "Dummy"


class BoxType(StrEnum):
    """Box types for the delivery service."""

    BASIC = "BASIC"
    GF = "GF"
    LA = "LA"
    VEGAN = "VEGAN"


class CellColors:
    """Colors for spreadsheet formatting."""

    BASIC: Final[str] = "00FFCC00"  # Orange
    HEADER: Final[str] = "00FFCCCC"  # Pink
    LA: Final[str] = "003399CC"  # Blue
    GF: Final[str] = "0099CC33"  # Green
    VEGAN: Final[str] = "00CCCCCC"  # Grey


BOX_TYPE_COLOR_MAP: Final[dict[str, str]] = {
    BoxType.BASIC: CellColors.BASIC,
    BoxType.GF: CellColors.GF,
    BoxType.LA: CellColors.LA,
    BoxType.VEGAN: CellColors.VEGAN,
}


class CircuitColumns:
    """Column/field/doc name constants for Circuit API."""

    ACTIVE: Final[str] = "active"
    ADDRESS: Final[str] = "address"
    ADDRESS_NAME: Final[str] = "addressName"
    ADDRESS_LINE_1: Final[str] = "addressLineOne"
    ADDRESS_LINE_2: Final[str] = "addressLineTwo"
    ALLOWED_DRIVERS: Final[str] = "allowedDrivers"
    CANCELED: Final[str] = "canceled"
    CITY: Final[str] = "city"
    CODE: Final[str] = "code"
    COUNTRY: Final[str] = "country"
    DAY: Final[str] = "day"
    DEPOT: Final[str] = "depot"
    DISTRIBUTED: Final[str] = "distributed"
    DONE: Final[str] = "done"
    DRIVERS: Final[str] = "drivers"
    EMAIL: Final[str] = "email"
    EXTERNAL_ID: Final[str] = "externalId"
    ID: Final[str] = "id"
    METADATA: Final[str] = "metadata"
    MONTH: Final[str] = "month"
    NAME: Final[str] = "name"
    NOTES: Final[str] = "notes"
    OPERATIONS: Final[str] = "operations"
    OPTIMIZATION: Final[str] = "optimization"
    ORDER_INFO: Final[str] = "orderInfo"
    PACKAGE_COUNT: Final[str] = "packageCount"
    PHONE: Final[str] = "phone"
    PLACE_ID: Final[str] = "placeId"
    PLAN: Final[str] = "plan"
    PLANS: Final[str] = "plans"
    PRODUCTS: Final[str] = "products"
    RECIPIENT: Final[str] = "recipient"
    RESULT: Final[str] = "result"
    ROUTE: Final[str] = "route"
    SKIPPED_STOPS: Final[str] = "skippedStops"
    STARTS: Final[str] = "starts"
    STATE: Final[str] = "state"
    STOP_POSITION: Final[str] = "stopPosition"
    STOPS: Final[str] = "stops"
    TITLE: Final[str] = "title"
    WRITABLE: Final[str] = "writable"
    YEAR: Final[str] = "year"
    ZIP: Final[str] = "zip"


CIRCUIT_URL: Final[str] = "https://api.getcircuit.com/public/v0.2b"
CIRCUIT_DRIVERS_URL: Final[str] = f"{CIRCUIT_URL}/drivers"


class Columns:
    """Column name constants."""

    ADDRESS: Final[str] = "Address"
    BOX_TYPE: Final[str] = "Box Type"
    BOX_COUNT: Final[str] = "Box Count"
    DRIVER: Final[str] = "Driver"
    EMAIL: Final[str] = "Email"
    NAME: Final[str] = "Name"
    NEIGHBORHOOD: Final[str] = "Neighborhood"
    NOTES: Final[str] = "Notes"
    ORDER_COUNT: Final[str] = "Order Count"
    PHONE: Final[str] = "Phone"
    PRODUCT_TYPE: Final[str] = "Product Type"
    STOP_NO: Final[str] = "Stop #"


CIRCUIT_DATE_FORMAT: Final[str] = "%Y-%m-%d"

COLUMN_NAME_MAP: Final[dict[str, str]] = {Columns.BOX_TYPE: Columns.PRODUCT_TYPE}


COMBINED_ROUTES_COLUMNS: Final[list[str]] = [
    Columns.STOP_NO,
    Columns.NAME,
    Columns.ADDRESS,
    Columns.PHONE,
    Columns.NOTES,
    Columns.ORDER_COUNT,
    Columns.BOX_TYPE,
    Columns.NEIGHBORHOOD,
]

CIRCUIT_DOWNLOAD_COLUMNS: Final[list[str]] = COMBINED_ROUTES_COLUMNS + [Columns.EMAIL]


# TODO: Combine with DocString?
# https://github.com/crickets-and-comb/bfb_delivery/issues/57
class Defaults:
    """Default values. E.g., for syncing public API with CLI."""

    BUILD_ROUTES_FROM_CHUNKED: Final[dict[str, str | bool]] = {
        "output_dir": "",
        "start_date": "",
        "no_distribute": False,
        "verbose": False,
        "book_one_drivers_file": "",
        "extra_notes_file": "",
    }
    COMBINE_ROUTE_TABLES: Final[dict[str, str]] = {"output_dir": "", "output_filename": ""}
    CREATE_MANIFESTS: Final[dict[str, str]] = {
        "output_dir": "",
        "output_filename": "",
        "extra_notes_file": "",
    }
    CREATE_MANIFESTS_FROM_CIRCUIT: Final[dict[str, str | bool | list]] = {
        "start_date": "",
        "end_date": "",
        "plan_ids": [],
        "output_dir": CREATE_MANIFESTS["output_dir"],
        "output_filename": CREATE_MANIFESTS["output_filename"],
        "circuit_output_dir": "",
        "all_hhs": False,
        "verbose": False,
        "extra_notes_file": CREATE_MANIFESTS["extra_notes_file"],
    }
    FORMAT_COMBINED_ROUTES: Final[dict[str, str]] = {
        "output_dir": "",
        "output_filename": "",
        "extra_notes_file": CREATE_MANIFESTS["extra_notes_file"],
    }
    SPLIT_CHUNKED_ROUTE: Final[dict[str, str | int]] = {
        "output_dir": "",
        "output_filename": "",
        "n_books": 4,
        "book_one_drivers_file": "",
        "date": "",
    }


class DocStrings:
    """Docstrings for the public API."""

    BUILD_ROUTES_FROM_CHUNKED: Final = DocString(
        opening="""
Build and disbribute routes from chunked routes.

From a chunked route spreadsheet, builds, optimizes, and distributes routes to drivers.
Produces a final manifest spreadsheet.

Prompts interactive user input to confirm driver assignments.

See :doc:`build_routes_from_chunked` for more information.
""",
        args={
            "input_path": "Path to the chunked route spreadsheet.",
            "output_dir": (
                "Path to the output directory. Empty defaults to a new directory in the "
                'present working directory, named "deliveries_{date}".'
            ),
            "start_date": (
                'The date to start the routes, as "YYYY-MM-DD". Empty string defaults to '
                "the soonest Friday."
            ),
            "no_distribute": "To skip distributing the routes to drivers after optimizing.",
            "verbose": "Whether to print verbose output.",
            "extra_notes_file": (
                "Path to the extra notes file. If empty, uses a constant DataFrame. "
                "See :py:data:`bfb_delivery.lib.constants.ExtraNotes`."
            ),
        },
        returns=["The path to the final manifest spreadsheet."],
        raises=[],
    )

    COMBINE_ROUTE_TABLES: Final = DocString(
        opening="""
Combines the driver route CSVs into a single workbook.

This is used after optimizing and exporting the routes to individual CSVs. It prepares the
worksheets to be formatted with :py:func:`bfb_delivery.api.public.format_combined_routes`.

If `output_dir` is specified, will create the directory if it doesn't exist.

.. note::

    Changes "Product Type" column name back to "Box Type".

See :doc:`combine_route_tables` for more information.
""",
        args={
            "input_dir": "The directory containing the driver route CSVs.",
            "output_dir": (
                "The directory to write the output workbook to. "
                "Empty string (default) saves to the `input_dir` directory."
            ),
            "output_filename": (
                "The name of the output workbook. "
                "Empty string (default) will name the file 'combined_routes_{date}.xlsx'."
            ),
        },
        raises=[
            ErrorDocString(error_type="ValueError", docstring="If `input_paths` is empty.")
        ],
        returns=["The path to the output workbook."],
    )

    CREATE_MANIFESTS: Final = DocString(
        opening="""
From Circuit route CSVs, creates driver manifest workbook ready to print.

This is used after optimizing and exporting the routes to individual CSVs. Reads in
driver route CSVs from `input_dir` and creates a formatted workbook with driver
manifests ready to print, with headers, aggregate data, and color-coded box types. Each
driver's route is a separate sheet in the workbook.

The workbook is saved to `output_dir` with the name `output_filename`. Will create
`output_dir` if it doesn't exist.

.. note::

    Uses the date of the front of each CSV name to set the manifest date field. I.e.,
    each sheet should be named something like "08.08 Richard N", and, e.g., this would
    set the manifest date field to "Date: 08.08".

Just wraps :py:func:`bfb_delivery.api.public.combine_route_tables` and
:py:func:`bfb_delivery.api.public.format_combined_routes`. Creates an intermediate output
workbook with all routes combined, then formats it.

See :doc:`create_manifests` for more information.
""",
        args={
            "input_dir": "The directory containing the driver route CSVs.",
            "output_dir": (
                "The directory to write the formatted manifest workbook to. "
                "Empty string (default) saves to the `input_dir` directory."
            ),
            "output_filename": (
                "The name of the output workbook."
                'Empty string sets filename to "final_manifests_{date}.xlsx".'
            ),
            "extra_notes_file": (
                "Path to the extra notes file. If empty (default), uses a constant "
                "DataFrame. See :py:data:`bfb_delivery.lib.constants.ExtraNotes`."
            ),
        },
        returns=["Path to the formatted manifest workbook."],
        raises=[],
    )

    CREATE_MANIFESTS_FROM_CIRCUIT: Final = DocString(
        opening="""
Gets optimized routes from Circuit, creates driver manifest workbook ready to print.

This is used after uploading and optimizing the routes. Reads routes CSVs from Circuit,
and creates a formatted workbook with driver manifests ready to print, with headers,
aggregate data, and color-coded box types. Each driver's route is a separate sheet in the
workbook.

The workbook is saved to `output_dir` with the name `output_filename`. Will create
`output_dir` if it doesn't exist.

.. note::

    Uses the date of the front of each CSV name to set the manifest date field. I.e.,
    each sheet should be named something like "08.08 Richard N", and, e.g., this would
    set the manifest date field to "Date: 08.08". **But, this does not determine the
    search date range.**

Wraps :py:func:`bfb_delivery.api.public.create_manifests` and adds Circuit integration.
And, `create_manifests` just wraps :py:func:`bfb_delivery.api.public.combine_route_tables`
and :py:func:`bfb_delivery.api.public.format_combined_routes`. Creates an intermediate
output workbook with all routes combined, then formats it.

See :doc:`create_manifests_from_circuit` for more information.
""",
        args={
            "start_date": (
                'The start date to use in the output workbook sheetnames as "YYYYMMDD". '
                "Empty string uses the soonest Friday. Range is inclusive."
            ),
            "end_date": (
                'The end date to use in the output workbook sheetnames as "YYYYMMDD". '
                "Empty string uses the start date. Range is inclusive."
            ),
            "plan_ids": (
                "The list of plan IDs to filter the Circuit routes by. Overrides `all_hhs`. "
                "Not valid for CLI."
            ),
            "output_dir": (
                "The directory to write the formatted manifest workbook to. "
                "Empty string saves to the `input_dir` directory."
            ),
            "output_filename": (
                "The name of the output workbook. "
                'Empty string sets filename to "final_manifests_{date}.xlsx".'
            ),
            "circuit_output_dir": (
                "The directory to create a subdir to save the routes to. Creates "
                '"routes_{date}" directory within the `circuit_output_dir`. Empty string '
                "uses `output_dir`. If the directory does not exist, it is created. If it "
                "exists, it is overwritten."
            ),
            "all_hhs": (
                'Flag to get only the "All HHs" route. '
                'False gets all routes except "All HHs". True gets only the "All HHs" route. '
                "Overridden by `plan_ids`."
            ),
            "verbose": "Flag to print verbose output.",
            "extra_notes_file": (
                "Path to the extra notes file. If empty, uses a constant "
                "DataFrame. See :py:data:`bfb_delivery.lib.constants.ExtraNotes`."
            ),
        },
        returns=["Path to the final manifest workbook."],
        raises=[],
    )

    FORMAT_COMBINED_ROUTES: Final = DocString(
        opening="""
Formats the combined routes table into driver manifests to print.

Adds headers and aggregate data. Color codes box types.

This is used after combining the driver route CSVs into a single workbook
using :py:func:`bfb_delivery.api.public.combine_route_tables`.

If `output_dir` is specified, will create the directory if it doesn't exist.

.. note::

    Uses the date of the front of each sheet name to set the manifest date field. I.e.,
    each sheet should be named something like "05.27 Oscar W", and, e.g., this would set
    the manifest date field to "Date: 05.27".

See :doc:`format_combined_routes` for more information.
""",
        args={
            "input_path": "The path to the combined routes table.",
            "output_dir": (
                "The directory to write the formatted table to. "
                "Empty string (default) saves to the input path's parent directory."
            ),
            "output_filename": (
                "The name of the formatted workbook. "
                'Empty string (default) will name the file "formatted_routes_{date}.xlsx".'
            ),
            "extra_notes_file": (
                "The path to the extra notes file. If empty (default), uses a "
                "constant DataFrame. See :py:data:`bfb_delivery.lib.constants.ExtraNotes`."
            ),
        },
        returns=["The path to the formatted table."],
        raises=[],
    )

    SPLIT_CHUNKED_ROUTE: Final = DocString(
        opening="""
Split route sheet into n workbooks with sheets by driver.

Sheets by driver allows splitting routes by driver on Circuit upload.
Multiple workbooks allows team to split the uploads among members, so one person
doesn't have to upload all routes.
This process follows the "chunking" process in the route generation, where routes
are split into smaller "chunks" by driver (i.e., each stop is labeled with a driver).

Reads a route spreadsheet at `input_path`.
Writes `n_books` Excel workbooks with each sheet containing the stops for a single driver.
Writes adjacent to the original workbook unless `output_dir` specified. If specified, will
create the directory if it doesn't exist.

.. note::

    Renames "Box Type" column name to "Product Type", per Circuit API.

.. note::

    The date passed sets the date in the sheet names of the output workbooks, and that
    date in the sheet name is used for the manifest date field in later functions that
    make the manifests: :py:func:`bfb_delivery.api.public.format_combined_routes` and
    :py:func:`bfb_delivery.api.public.create_manifests_from_circuit` (which wraps the former).


See :doc:`split_chunked_route` for more information.
""",
        args={
            "input_path": (
                "Path to the chunked route sheet that this function reads in and "
                "splits up."
            ),
            "output_dir": (
                "Directory to save the output workbook. "
                "Empty string saves to the input `input_path` directory."
            ),
            "output_filename": (
                "Name of the output workbook. "
                'Empty string sets filename to "split_workbook_{date}_{i of n_books}.xlsx".'
            ),
            "n_books": "Number of workbooks to split into.",
            # TODO: Standardize some of these definitions as well.
            # https://github.com/crickets-and-comb/bfb_delivery/issues/58
            "book_one_drivers_file": (
                "Path to the book-one driver's file. If empty, uses a constant list. "
                "See :py:data:`bfb_delivery.lib.constants.BookOneDrivers`."
            ),
            "date": (
                "The date to use in the output workbook sheetnames. Empty string (default) "
                "uses the soonest Friday."
            ),
        },
        returns=["Paths to the split chunked route workbooks."],
        raises=[
            ErrorDocString(error_type="ValueError", docstring="If `n_books` is less than 1."),
            ErrorDocString(
                error_type="ValueError",
                docstring=(
                    "If `n_books` is greater than the number of drivers in the input "
                    "workbook."
                ),
            ),
        ],
    )


# Food placeId.
DEPOT_PLACE_ID: Final[str] = "ChIJFw9CDZejhVQRizqiyJSmPqo"


class ExtraNotes:
    """Extra notes to add at bottom of manifests when any stop is tagged.

    Is a class so it appears in docs.
    """

    #: Tuples of tag and note. Example: ("Cascade Meadows Apartments*", "Note here.")
    notes: Final[list[tuple[str, str]]] = [
        # ("Cascade Meadows Apartments*", ""),
        # ("Deer Run Terrace Apartments*", ""),
        # ("Eleanor Apartments*", ""),
        # ("Evergreen Ridge Apartments*", ""),
        # ("Gardenview Village*", ""),
        # ("Heart House*", ""),
        # ("Laurel Forest Apartments*", ""),
        # ("Laurel Village*", ""),
        # ("Park Ridge Apartments*", ""),
        # ("Regency Park Apartments*", ""),
        # ("Sterling Senior Apartments*", ""),
        # ("Trailview Apartments*", ""),
        # ("Tullwood Apartments*", ""),
        # ("Varsity Village*", ""),
        # ("Walton Place*", ""),
        # ("Washington Square Apartments*", ""),
        # ("Woodrose Apartments*", ""),
        # ("Washington Grocery Building*", ""),
    ]

    #: Extra notes DataFrame. The columns are "tag" and "note". Made from `notes`.
    df: Final[pd.DataFrame]

    def __init__(self) -> None:
        """Initialize the extra notes `df`."""
        self.df = pd.DataFrame(columns=["tag", "note"], data=self.notes)


FILE_DATE_FORMAT: Final[str] = "%Y%m%d"

FORMATTED_ROUTES_COLUMNS: Final[list[str]] = [
    Columns.STOP_NO,
    Columns.NAME,
    Columns.ADDRESS,
    Columns.PHONE,
    Columns.NOTES,
    Columns.BOX_TYPE,
]


class IntermediateColumns:
    """Column name constants for intermediate tables."""

    DRIVER_NAME: Final[str] = "driver_name"
    DRIVER_SHEET_NAME: Final[str] = "driver_sheet_name"
    INITIALIZED: Final[str] = "initialized"
    OPTIMIZED: Final[str] = "optimized"
    PLAN_ID: Final[str] = "plan_id"
    ROUTE_TITLE: Final[str] = "route_title"  # AKA plan title, since 1:1.
    SHEET_NAME: Final[str] = "sheet_name"
    START_DATE: Final[str] = "start_date"
    STOPS_UPLOADED: Final[str] = "stops_uploaded"


MANIFEST_DATE_FORMAT: Final[str] = "%m.%d"

MAX_ORDER_COUNT: Final[int] = 5

NOTES_COLUMN_WIDTH: Final[float] = 56.67

PROTEIN_BOX_TYPES: Final[list[str]] = ["BASIC", "GF", "LA"]


class RateLimits:
    """Rate limits for Circuit API."""

    BATCH_STOP_IMPORT_SECONDS: Final[float] = 6
    BATCH_STOP_IMPORT_MAX_STOPS: Final[int] = 1000
    OPTIMIZATION_PER_SECOND: Final[float] = 20
    OPTIMIZATION_TIMEOUT_SECONDS: Final[float] = 10
    READ_TIMEOUT_SECONDS: Final[float] = OPTIMIZATION_TIMEOUT_SECONDS
    READ_SECONDS: Final[float] = 0.1
    WAIT_DECREASE_SECONDS: Final[float] = 0.6
    WAIT_INCREASE_SCALAR: Final[float] = 2
    WRITE_SECONDS: Final[float] = 0.2
    WRITE_TIMEOUT_SECONDS: Final[float] = OPTIMIZATION_TIMEOUT_SECONDS


SPLIT_ROUTE_COLUMNS: Final[list[str]] = [
    Columns.NAME,
    Columns.ADDRESS,
    Columns.PHONE,
    Columns.EMAIL,
    Columns.NOTES,
    Columns.ORDER_COUNT,
    Columns.PRODUCT_TYPE,
    Columns.NEIGHBORHOOD,
]



# noqa: D100
__doc__ = """
.. click:: bfb_delivery.cli.build_routes_from_chunked:main
    :prog: build_routes_from_chunked
    :nested: full
"""

import logging

import click
from typeguard import typechecked

from bfb_delivery import build_routes_from_chunked
from bfb_delivery.lib.constants import Defaults, DocStrings

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@click.command(help=DocStrings.BUILD_ROUTES_FROM_CHUNKED.cli_docstring)
@click.option(
    "--input_path",
    type=str,
    required=True,
    help=DocStrings.BUILD_ROUTES_FROM_CHUNKED.args["input_path"],
)
@click.option(
    "--output_dir",
    type=str,
    required=False,
    default=Defaults.BUILD_ROUTES_FROM_CHUNKED["output_dir"],
    help=DocStrings.BUILD_ROUTES_FROM_CHUNKED.args["output_dir"],
)
@click.option(
    "--start_date",
    type=str,
    required=False,
    default=Defaults.BUILD_ROUTES_FROM_CHUNKED["start_date"],
    help=DocStrings.BUILD_ROUTES_FROM_CHUNKED.args["start_date"],
)
@click.option(
    "--no_distribute",
    is_flag=True,
    default=Defaults.BUILD_ROUTES_FROM_CHUNKED["no_distribute"],
    help=DocStrings.BUILD_ROUTES_FROM_CHUNKED.args["no_distribute"],
)
@click.option(
    "--verbose",
    is_flag=True,
    default=Defaults.BUILD_ROUTES_FROM_CHUNKED["verbose"],
    help=DocStrings.BUILD_ROUTES_FROM_CHUNKED.args["verbose"],
)
@click.option(
    "--extra_notes_file",
    type=str,
    required=False,
    default=Defaults.BUILD_ROUTES_FROM_CHUNKED["extra_notes_file"],
    help=DocStrings.BUILD_ROUTES_FROM_CHUNKED.args["extra_notes_file"],
)
@typechecked
def main(  # noqa: D103
    input_path: str,
    output_dir: str,
    start_date: str,
    no_distribute: bool,
    verbose: bool,
    extra_notes_file: str,
) -> str:
    final_manifest_path = build_routes_from_chunked(
        input_path=input_path,
        output_dir=output_dir,
        start_date=start_date,
        no_distribute=no_distribute,
        verbose=verbose,
        extra_notes_file=extra_notes_file,
    )
    logger.info(f"Formatted workbook saved to:\n{final_manifest_path.resolve()}")
    # Print statement to capture in tests.
    print(str(final_manifest_path))

    return str(final_manifest_path)



"""Utility functions for schema validation."""

from collections.abc import Callable

from pandera.errors import SchemaError, SchemaErrors


def schema_error_handler(func: Callable) -> Callable:
    """Custom error handler for the schema validation errors.

    Use as decorator to wrap pa.check_types decorator.
    """

    def wrapper(*args, **kwargs):  # noqa: ANN201, ANN002, ANN003
        try:
            return func(*args, **kwargs)
        except SchemaError as e:
            # If error message too verbose, hop in here.
            # Or, if specifying the message for testing.
            # e_dict = vars(e)
            # err_msg = "SchemaError: "
            # schema = e_dict.get("schema")
            # reason_code = e_dict.get("reason_code")
            # column_name = e_dict.get("column_name")
            # check = e_dict.get("check")
            # failure_cases = e_dict.get("failure_cases")
            # if schema:
            #     err_msg += f"\nSchema: {schema}"
            # if reason_code:
            #     err_msg += f"\nReason code: {reason_code}"
            # if column_name:
            #     err_msg += f"\nColumn name: {column_name}"
            # if check:
            #     err_msg += f"\nCheck: {check}"

            # raise SchemaError(schema=schema, data=failure_cases, message=err_msg) from e
            raise e
        except SchemaErrors as e:
            # If lazy evaluation is too verbose, hop in here.
            # Or, if specifying the message for testing.
            raise e
        except Exception as e:
            raise e

    return wrapper



"""The data schema for ETL steps."""

from functools import partial
from typing import Any

import pandera as pa
from pandera.typing import Series

from bfb_delivery.lib.constants import (
    DEPOT_PLACE_ID,
    BoxType,
    CircuitColumns,
    Columns,
    IntermediateColumns,
)
from bfb_delivery.lib.schema import checks  # noqa: F401

_COERCE_FIELD = partial(pa.Field, coerce=True)
_NULLABLE_FIELD = partial(_COERCE_FIELD, nullable=True)
_UNIQUE_FIELD = partial(_COERCE_FIELD, unique=True)

ADDRESS_FIELD = partial(_COERCE_FIELD, alias=Columns.ADDRESS)
BOX_TYPE_FIELD = partial(
    _COERCE_FIELD, in_list_case_insensitive={"category_list": BoxType}, alias=Columns.BOX_TYPE
)
DRIVER_ID_FIELD = partial(_COERCE_FIELD, str_startswith="drivers/")
# Renamed CircuitColumns.TITLE column, e.g. "1.17 Andy W":
EMAIL_FIELD = partial(_NULLABLE_FIELD, alias=Columns.EMAIL)
NAME_FIELD = partial(_COERCE_FIELD, alias=Columns.NAME)
NEIGHBORHOOD_FIELD = partial(_NULLABLE_FIELD, alias=Columns.NEIGHBORHOOD)
NOTES_FIELD = partial(_NULLABLE_FIELD, alias=Columns.NOTES)
ORDER_COUNT_FIELD = partial(_COERCE_FIELD, eq=1, alias=Columns.ORDER_COUNT)
ORDER_INFO_FIELD = partial(
    _COERCE_FIELD, item_in_field_dict=CircuitColumns.PRODUCTS, alias=CircuitColumns.ORDER_INFO
)
PHONE_FIELD = partial(_NULLABLE_FIELD, alias=Columns.PHONE)
# plan id e.g. "plans/0IWNayD8NEkvD5fQe2SQ":
PLAN_ID_FIELD = partial(_COERCE_FIELD, str_startswith="plans/")
ROUTE_FIELD = partial(_COERCE_FIELD, alias=CircuitColumns.ROUTE)
# stop id e.g. "plans/0IWNayD8NEkvD5fQe2SQ/stops/40lmbcQrd32NOfZiiC1b":
STOP_ID_FIELD = partial(
    _UNIQUE_FIELD, str_startswith="plans/", str_contains="/stops/", alias=CircuitColumns.ID
)
STOP_NO_FIELD = partial(_COERCE_FIELD, ge=1, alias=Columns.STOP_NO)
TITLE_FIELD = partial(_COERCE_FIELD, at_least_two_words=True)

# TODO: Unsmurf.
# https://github.com/crickets-and-comb/bfb_delivery/issues/83


class CircuitPlansOut(pa.DataFrameModel):
    """The schema for the Circuit plans data.

    bfb_delivery.lib.dispatch.read_circuit._make_plans_df output.
    """

    # plan id e.g. "plans/0IWNayD8NEkvD5fQe2SQ":
    id: Series[str] = PLAN_ID_FIELD(unique=True, alias=CircuitColumns.ID)
    # e.g. "1.17 Andy W":
    title: Series[str] = TITLE_FIELD(alias=CircuitColumns.TITLE)

    class Config:
        """The configuration for the schema."""

        strict = True


class CircuitPlansFromDict(CircuitPlansOut):
    """The schema for the Circuit plans data from a JSON-esque dict.

    bfb_delivery.lib.dispatch.read_circuit._make_plans_df input.
    """

    routes: Series[list[str]] = _COERCE_FIELD(is_list_of_one_or_less=True)

    class Config:
        """The configuration for the schema."""

        from_format = "dict"
        strict = False


class CircuitPlansTransformIn(CircuitPlansOut):
    """The schema for the Circuit plans data.

    bfb_delivery.lib.dispatch.read_circuit._transform_routes_df input.
    """


class CircuitRoutesTransformInFromDict(pa.DataFrameModel):
    """The schema for the Circuit routes data from a JSON-esque dict.

    bfb_delivery.lib.dispatch.read_circuit._transform_routes_df input.
    """

    plan: Series[str] = PLAN_ID_FIELD(alias=CircuitColumns.PLAN)
    route: Series[dict[str, Any]] = ROUTE_FIELD(item_in_field_dict=CircuitColumns.ID)
    id: Series[str] = STOP_ID_FIELD()
    # Position 0 is depot, which gets dropped later for the manifests.
    stopPosition: Series[int] = STOP_NO_FIELD(ge=0, alias=CircuitColumns.STOP_POSITION)
    recipient: Series[dict[str, Any]] = _COERCE_FIELD(
        item_in_field_dict=CircuitColumns.NAME, alias=CircuitColumns.RECIPIENT
    )
    address: Series[dict[str, Any]] = ADDRESS_FIELD(
        item_in_field_dict=CircuitColumns.PLACE_ID, alias=CircuitColumns.ADDRESS
    )
    notes: Series[str] = NOTES_FIELD(alias=CircuitColumns.NOTES)
    orderInfo: Series[dict[str, Any]] = ORDER_INFO_FIELD()
    packageCount: Series[float] = _NULLABLE_FIELD(eq=1, alias=CircuitColumns.PACKAGE_COUNT)

    class Config:
        """The configuration for the schema."""

        from_format = "dict"

        many_to_one = {"many_col": CircuitColumns.ID, "one_col": CircuitColumns.PLAN}
        unique_group = {
            "group_col": CircuitColumns.PLAN,
            "unique_col": CircuitColumns.STOP_POSITION,
        }
        contiguous_group = {
            "group_col": CircuitColumns.PLAN,
            "contiguous_col": CircuitColumns.STOP_POSITION,
            "start_idx": 0,
        }
        # TODO: Convert back to field checks now that we know they must be uniquely defined
        # either way.
        # https://github.com/crickets-and-comb/bfb_delivery/issues/84
        address1_in_address = True
        address2_in_address = True


class CircuitRoutesTransformOut(pa.DataFrameModel):
    """The schema for the Circuit routes data after transformation.

    bfb_delivery.lib.dispatch.read_circuit._transform_routes_df output.
    """

    # Main output columns for downstream processing.
    # route id e.g. "routes/lITTnQsxYffqJQDxIpzr".
    route: Series[str] = ROUTE_FIELD(str_startswith="routes/")
    driver_sheet_name: Series[str] = TITLE_FIELD(alias=IntermediateColumns.DRIVER_SHEET_NAME)
    stop_no: Series[int] = STOP_NO_FIELD()
    name: Series[str] = NAME_FIELD()
    address: Series[str] = ADDRESS_FIELD()
    phone: Series[str] = PHONE_FIELD()
    notes: Series[str] = NOTES_FIELD()
    order_count: Series[float] = ORDER_COUNT_FIELD()
    box_type: Series[pa.Category] = BOX_TYPE_FIELD()
    neighborhood: Series[str] = NEIGHBORHOOD_FIELD()
    email: Series[str] = EMAIL_FIELD()

    # Ancillary columns.
    plan: Series[str] = PLAN_ID_FIELD(alias=CircuitColumns.PLAN)
    id: Series[str] = STOP_ID_FIELD()
    orderInfo: Series[dict[str, Any]] = ORDER_INFO_FIELD(one_product=True)
    route_title: Series[str] = TITLE_FIELD(alias=IntermediateColumns.ROUTE_TITLE)
    placeId: Series[str] = _COERCE_FIELD(ne=DEPOT_PLACE_ID, alias=CircuitColumns.PLACE_ID)

    class Config:
        """The configuration for the schema."""

        # NOTE, the Circuit plan:route relationship is 1:m, but we only ever want 1:1.
        one_to_one = {"col_a": CircuitColumns.ROUTE, "col_b": CircuitColumns.PLAN}
        one_to_one_route_sheet = True
        at_least_one_in_group = {
            "group_col": CircuitColumns.PLAN,
            "at_least_one_col": IntermediateColumns.DRIVER_SHEET_NAME,
        }
        at_least_one_in_group_sheet_plan = True
        at_least_one_in_group_route_sheet = True
        at_least_one_in_group_sheet_route = True

        # TODO: Was violated on 10/4. Investigate, but ignore for now.
        # plans/jEvjLs3ViQkKPBcJVduF, routes/z9AmJkUnuQXUGHGsoxyG
        # Had route title "10.11 Sara" and driver sheet name (plan title) "10.4 Sara"
        # equal_cols = {
        #     "col_a": IntermediateColumns.ROUTE_TITLE,
        #     "col_b": IntermediateColumns.DRIVER_SHEET_NAME,
        # }
        # https://github.com/crickets-and-comb/bfb_delivery/issues/85

        many_to_one = {
            "many_col": CircuitColumns.ID,
            "one_col": IntermediateColumns.DRIVER_SHEET_NAME,
        }
        unique = [IntermediateColumns.DRIVER_SHEET_NAME, Columns.STOP_NO]
        at_least_one_in_group_sheet_stop = True
        contiguous_group = {
            "group_col": IntermediateColumns.DRIVER_SHEET_NAME,
            "contiguous_col": Columns.STOP_NO,
            "start_idx": 1,
        }
        increasing_by = {"cols": [IntermediateColumns.DRIVER_SHEET_NAME, Columns.STOP_NO]}


class CircuitRoutesWriteIn(pa.DataFrameModel):
    """The schema for the Circuit routes data before writing.

    bfb_delivery.lib.dispatch.read_circuit._write_routes_df input.
    """

    driver_sheet_name: Series[str] = TITLE_FIELD(alias=IntermediateColumns.DRIVER_SHEET_NAME)
    stop_no: Series[int] = STOP_NO_FIELD()
    name: Series[str] = NAME_FIELD()
    address: Series[str] = ADDRESS_FIELD()
    phone: Series[str] = PHONE_FIELD()
    notes: Series[str] = NOTES_FIELD()
    order_count: Series[float] = ORDER_COUNT_FIELD()
    box_type: Series[pa.Category] = BOX_TYPE_FIELD()
    neighborhood: Series[str] = NEIGHBORHOOD_FIELD()
    email: Series[str] | None = EMAIL_FIELD()

    class Config:
        """The configuration for the schema."""

        one_to_one = {
            "col_a": CircuitColumns.ROUTE,
            "col_b": IntermediateColumns.DRIVER_SHEET_NAME,
        }
        many_to_one = {
            "many_col": CircuitColumns.ID,
            "one_col": IntermediateColumns.DRIVER_SHEET_NAME,
        }
        at_least_one_in_group_route_sheet = True
        at_least_one_in_group_sheet_route = True

        unique = [IntermediateColumns.DRIVER_SHEET_NAME, Columns.STOP_NO]
        at_least_one_in_group_sheet_stop = True
        contiguous_group = {
            "group_col": IntermediateColumns.DRIVER_SHEET_NAME,
            "contiguous_col": Columns.STOP_NO,
            "start_idx": 1,
        }
        increasing_by = {"cols": [IntermediateColumns.DRIVER_SHEET_NAME, Columns.STOP_NO]}


class CircuitRoutesWriteInAllHHs(CircuitRoutesWriteIn):
    """The schema for the Circuit routes data before writing for "All HHs".

    bfb_delivery.lib.dispatch.read_circuit._write_routes_df_all_hhs input.
    """

    email: Series[str] = EMAIL_FIELD()

    class Config:
        """The configuration for the schema."""

        many_to_one = {
            "many_col": Columns.STOP_NO,
            "one_col": IntermediateColumns.DRIVER_SHEET_NAME,
        }


class CircuitRoutesWriteOut(pa.DataFrameModel):
    """The schema for the Circuit routes data after writing.

    bfb_delivery.lib.dispatch.read_circuit._write_routes_df input,
    called within _write_routes_df as its "output."
    """

    stop_no: Series[int] = pa.Field(
        coerce=True, unique=True, ge=1, contiguous=1, is_sorted=True, alias=Columns.STOP_NO
    )
    name: Series[str] = NAME_FIELD()
    address: Series[str] = ADDRESS_FIELD()
    phone: Series[str] = PHONE_FIELD()
    notes: Series[str] = NOTES_FIELD()
    order_count: Series[float] = ORDER_COUNT_FIELD()
    box_type: Series[pa.Category] = BOX_TYPE_FIELD()
    neighborhood: Series[str] = NEIGHBORHOOD_FIELD()
    email: Series[str] = EMAIL_FIELD()

    class Config:
        """The configuration for the schema."""

        unique = [Columns.NAME, Columns.ADDRESS, Columns.BOX_TYPE]


class Stops(pa.DataFrameModel):
    """The schema for the stops data to upload."""

    name: Series[str] = NAME_FIELD()
    address: Series[str] = ADDRESS_FIELD()
    phone: Series[str] = PHONE_FIELD()
    email: Series[str] = EMAIL_FIELD()
    notes: Series[str] = NOTES_FIELD()
    order_count: Series[float] = ORDER_COUNT_FIELD()
    product_type: Series[pa.Category] = BOX_TYPE_FIELD(alias=Columns.PRODUCT_TYPE)
    neighborhood: Series[str] = NEIGHBORHOOD_FIELD()
    sheet_name: Series[str] = TITLE_FIELD(alias=IntermediateColumns.SHEET_NAME)

    class Config:
        """The configuration for the schema."""

        unique = [Columns.NAME, Columns.ADDRESS, Columns.PRODUCT_TYPE]


class DriversGetAllDriversOut(pa.DataFrameModel):
    """The schema for the drivers data after getting all drivers."""

    id: Series[str] = DRIVER_ID_FIELD(unique=True, alias=CircuitColumns.ID)
    name: Series[str] = NAME_FIELD(alias=CircuitColumns.NAME)
    email: Series[str] = EMAIL_FIELD(nullable=False, alias=CircuitColumns.EMAIL)
    active: Series[bool] = _COERCE_FIELD(alias=CircuitColumns.ACTIVE)


class DriversAssignDriversIn(DriversGetAllDriversOut):
    """The schema for the drivers data before assigning drivers."""


class DriversAssignDriverIn(DriversAssignDriversIn):
    """The schema for the driver data before assigning a driver."""


class PlansAssignDriversIn(pa.DataFrameModel):
    """The schema for the plans data before assigning drivers."""

    route_title: Series[str] = TITLE_FIELD(unique=True, alias=IntermediateColumns.ROUTE_TITLE)
    driver_name: Series[str] = NAME_FIELD(
        nullable=True, alias=IntermediateColumns.DRIVER_NAME
    )
    email: Series[str] = EMAIL_FIELD(alias=CircuitColumns.EMAIL)
    id: Series[str] = DRIVER_ID_FIELD(nullable=True, alias=CircuitColumns.ID)


class PlansAssignDriverIn(PlansAssignDriversIn):
    """The schema for the plan data before assigning a driver."""


class PlansAssignDriverOut(PlansAssignDriverIn):
    """The schema for the plan data after assigning a driver."""


class PlansAssignDriversOut(PlansAssignDriverOut):
    """The schema for the plans data after assigning drivers."""


class PlansAssignDriversToPlansOut(PlansAssignDriversOut):
    """The schema for the plans data after assigning drivers."""

    driver_name: Series[str] = NAME_FIELD(alias=IntermediateColumns.DRIVER_NAME)
    email: Series[str] = EMAIL_FIELD(nullable=False, alias=CircuitColumns.EMAIL)
    id: Series[str] = DRIVER_ID_FIELD(alias=CircuitColumns.ID)

    class Config:
        """The configuration for the schema."""

        many_to_one = {
            "many_col": IntermediateColumns.ROUTE_TITLE,
            "one_col": CircuitColumns.ID,  # Driver ID
        }


class PlansInitializePlansIn(PlansAssignDriversToPlansOut):
    """The schema for the plans data before initializing."""


class PlansInitializePlansOut(PlansInitializePlansIn):
    """The schema for the plans data after initializing."""

    plan_id: Series[str] = PLAN_ID_FIELD(
        unique=True, nullable=True, alias=IntermediateColumns.PLAN_ID
    )
    writable: Series[bool] = _NULLABLE_FIELD(alias=CircuitColumns.WRITABLE)
    optimization: Series[str] = _NULLABLE_FIELD(alias=CircuitColumns.OPTIMIZATION)
    initialized: Series[bool] = _COERCE_FIELD(alias=IntermediateColumns.INITIALIZED)

    class Config:
        """The configuration for the schema."""

        # TODO: Make sure this accommodates null plan IDs.
        # https://github.com/crickets-and-comb/bfb_delivery/issues/86
        one_to_one = {
            "col_a": IntermediateColumns.ROUTE_TITLE,
            "col_b": IntermediateColumns.PLAN_ID,
        }


class PlansCreatePlansOut(PlansInitializePlansOut):
    """The schema for the plans data after creating."""


class PlansUploadStopsIn(PlansCreatePlansOut):
    """The schema for the plans data before uploading stops."""


class PlansBuildStopsIn(PlansCreatePlansOut):
    """The schema for the plans data before building stops."""

    # TODO: Refactor to relax the requirements, here and elsewhere.
    # https://github.com/crickets-and-comb/bfb_delivery/issues/87


class PlansUploadStopsOut(PlansUploadStopsIn):
    """The schema for the plan data after uploading stops."""

    stops_uploaded: Series[bool] = _COERCE_FIELD(alias=IntermediateColumns.STOPS_UPLOADED)


# TODO: Build a from/to dict validator for _build_stop_array?
# https://github.com/crickets-and-comb/bfb_delivery/issues/88


class PlansOptimizeRoutesIn(PlansUploadStopsOut):
    """The schema for the plans data before optimizing routes."""


class PlansConfirmOptimizationsIn(PlansOptimizeRoutesIn):
    """The schema for the plans data after before confirming optimizations."""

    routes_optimized: Series[bool] = _COERCE_FIELD(alias=IntermediateColumns.OPTIMIZED)


class PlansConfirmOptimizationsOut(PlansConfirmOptimizationsIn):
    """The schema for the plans data after confirming optimizations."""

    routes_optimized: Series = pa.Field(
        nullable=True, coerce=False, alias=IntermediateColumns.OPTIMIZED
    )


class PlansOptimizeRoutesOut(PlansConfirmOptimizationsOut):
    """The schema for the plans data after optimizing routes."""


class PlansDistributeRoutesIn(PlansOptimizeRoutesOut):
    """The schema for the plans data before distributing routes."""


class PlansDistributeRoutesOut(PlansDistributeRoutesIn):
    """The schema for the plans data after distributing routes."""

    distributed: Series[bool] = _COERCE_FIELD(alias=CircuitColumns.DISTRIBUTED)


class PlansUploadSplitChunkedOut(PlansDistributeRoutesOut):
    """The schema for the plans data after uploading split chunked routes."""
